import numpy as np
import pytest
from tdastro.utils.io_utils import read_grid_data, read_lclib_data


def test_read_grid_data_good(grid_data_good_file):
    """Test that we can read a well formatted grid data file."""
    x0, x1, values = read_grid_data(grid_data_good_file, format="ascii.csv")
    x0_expected = np.array([0.0, 1.0, 2.0])
    x1_expected = np.array([1.0, 1.5])
    values_expected = np.array([[0.0, 1.0], [2.0, 3.0], [4.0, 5.0]])

    np.testing.assert_allclose(x0, x0_expected, atol=1e-5)
    np.testing.assert_allclose(x1, x1_expected, atol=1e-5)
    np.testing.assert_allclose(values, values_expected, atol=1e-5)


def test_read_grid_data_bad(grid_data_bad_file):
    """Test that we correctly handle a badly formatted grid data file."""
    # We load without a problem is validation is off.
    x0, x1, values = read_grid_data(grid_data_bad_file, format="ascii")
    assert values.shape == (3, 2)

    with pytest.raises(ValueError):
        _, _, _ = read_grid_data(grid_data_bad_file, format="ascii", validate=True)

    # We fail when loading a nonexistent file.
    with pytest.raises(FileNotFoundError):
        _, _, _ = read_grid_data("no_such_file_here", format="ascii", validate=True)


def test_read_lclib_data(test_data_dir):
    """Test reading a SNANA LCLIB data from a text file."""
    lc_file = test_data_dir / "test_lclib_data.TEXT"
    curves = read_lclib_data(lc_file)
    assert len(curves) == 3

    expected_cols = ["type", "time", "u", "g", "r", "i", "z"]
    expected_len = [20, 20, 15]
    expected_param = [
        {"TYPE": "1", "OTHER": "1"},
        {"TYPE": "1", "OTHER": "2"},
        {"TYPE": "6", "OTHER": "7"},
    ]
    for idx, curve in enumerate(curves):
        assert len(curve) == expected_len[idx]
        assert int(curve.meta["id"]) == idx
        assert curve.meta["RECUR_CLASS"] == "RECUR-PERIODIC"
        for key, value in expected_param[idx].items():
            assert curve.meta["PARVAL"][key] == value

        # We did not pick up anything in the documentation block.
        assert "PURPOSE" not in curve.meta

        # Check that the expected columns are present and type is "S" or "T"
        for col in expected_cols:
            assert col in curve.colnames
        assert np.all((curve["type"].data == "S") | (curve["type"].data == "T"))



import numpy as np
from astropy import units as u
from astropy.modeling.physical_models import BlackBody


def black_body_luminosity_density_per_solid(temperature, radius, wavelengths):
    """Calculate the black-body luminosity density per solid angle.

    It is L_nu / (4 pi) for a spherical isotropic source.

    Parameters
    ----------
    temperature : float
        The effective temperature of the star, in kelvins.
    radius : float
        The radius of the star, in cm.
    wavelengths : numpy.ndarray
        A length N array of wavelengths, in cm.

    Returns
    -------
    luminosity_density : numpy.ndarray
        A length N array of luminosity density values.
        Output is in CGS units of erg/s/Hz/steradian.

    Notes
    -----
    tdastro adopts nJy as the unit of flux density, so the output of this
    function may need to be converted to compatible units,
    e.g. multiply by 10^32.
    """
    black_body = BlackBody(temperature * u.K)
    # Convert intensity to units compatible with nJy flux density
    intensity_per_freq = black_body(wavelengths * u.cm).to_value(u.erg / u.s / u.cm**2 / u.Hz / u.steradian)
    # Integral over solid angle, on surface of sphere
    surface_flux = intensity_per_freq * np.pi
    # Multiply to total area (4pi r^2) and divide by 4pi steradians,
    # e.g. multiply by r^2.
    surface_area_per_solid_angle = radius**2
    return surface_flux * surface_area_per_solid_angle



import os

import numpy as np
from tdastro.sources.spline_model import SplineModel
from tdastro.utils.wave_extrapolate import ConstantExtrapolation


def test_spline_model_flat() -> None:
    """Test that we can sample and create a flat SplineModel object."""
    times = np.linspace(1.0, 5.0, 20)
    wavelengths = np.linspace(100.0, 500.0, 25)
    fluxes = np.full((len(times), len(wavelengths)), 1.0)
    model = SplineModel(times, wavelengths, fluxes)
    assert str(model) == "SplineModel"

    test_times = np.array([0.0, 1.0, 2.0, 3.0, 10.0])
    test_waves = np.array([0.0, 100.0, 150.0, 200.0, 500.0, 1000.0])

    state = model.sample_parameters()
    values = model.evaluate(test_times, test_waves, state)
    assert values.shape == (5, 6)

    # The first and last times are outside the range of the model, so they should be 0.0
    # and the rest should be 1.0.
    expected = np.full_like(values, 1.0)
    expected[:, 0] = 0.0
    expected[:, -1] = 0.0
    np.testing.assert_array_almost_equal(values, expected)

    model2 = SplineModel(times, wavelengths, fluxes, amplitude=5.0, node_label="test")
    assert str(model2) == "test"

    state2 = model2.sample_parameters()
    values2 = model2.evaluate(test_times, test_waves, state2)
    assert values2.shape == (5, 6)

    # The first and last times are outside the range of the model, so they should be 0.0
    # and the rest should be 5.0.
    expected2 = np.full_like(values2, 5.0)
    expected2[:, 0] = 0.0
    expected2[:, -1] = 0.0
    np.testing.assert_array_almost_equal(values2, expected2)


def test_spline_model_interesting() -> None:
    """Test that we can sample and create a flat SplineModel object."""
    times = np.array([1.0, 2.0, 3.0])
    wavelengths = np.array([10.0, 20.0, 30.0])
    fluxes = np.array([[1.0, 5.0, 1.0], [5.0, 10.0, 5.0], [1.0, 5.0, 3.0]])
    model = SplineModel(
        times,
        wavelengths,
        fluxes,
        time_degree=1,
        wave_degree=1,
        wave_extrapolation=ConstantExtrapolation(value=0.1),
    )
    state = model.sample_parameters()

    test_times = np.array([1.0, 1.5, 2.0, 3.0])
    test_waves = np.array([10.0, 15.0, 20.0, 30.0, 40.0, 50.0])
    values = model.evaluate(test_times, test_waves, state)
    assert values.shape == (4, 6)

    # The last two wavelengths are outside the range of the model, so they should be 0.1
    expected = np.array(
        [
            [1.0, 3.0, 5.0, 1.0, 0.1, 0.1],
            [3.0, 5.25, 7.5, 3.0, 0.1, 0.1],
            [5.0, 7.5, 10.0, 5.0, 0.1, 0.1],
            [1.0, 3.0, 5.0, 3.0, 0.1, 0.1],
        ]
    )
    np.testing.assert_array_almost_equal(values, expected)


def test_spline_model_interesting_t0() -> None:
    """Test that we can sample and create a flat SplineModel object."""
    t0_times = np.array([1.0, 2.0, 3.0])
    wavelengths = np.array([10.0, 20.0, 30.0])
    fluxes = np.array([[1.0, 5.0, 1.0], [5.0, 10.0, 5.0], [1.0, 5.0, 3.0]])
    model = SplineModel(
        t0_times,
        wavelengths,
        fluxes,
        time_degree=1,
        wave_degree=1,
        t0=60676.0,
    )
    state = model.sample_parameters()

    # Test times correspond to t0 + [1.0, 1.5, 2.0, 3.0]
    test_times = np.array([60677.0, 60677.5, 60678.0, 60679.0])
    test_waves = np.array([10.0, 15.0, 20.0, 30.0])
    values = model.evaluate(test_times, test_waves, state)
    assert values.shape == (4, 4)

    expected = np.array(
        [[1.0, 3.0, 5.0, 1.0], [3.0, 5.25, 7.5, 3.0], [5.0, 7.5, 10.0, 5.0], [1.0, 3.0, 5.0, 3.0]]
    )
    np.testing.assert_array_almost_equal(values, expected)


def test_spine_model_from_file(test_data_dir):
    """Test that we can create a SplineModel from a file."""
    filename = os.path.join(test_data_dir, "truncated-salt2-h17/salt2_template_0.dat")

    model = SplineModel.from_file(filename)
    assert len(model._times) == 26
    assert len(model._wavelengths) == 401



"""A class that can be used to quickly sample from a one-dimensional distribution provided
as an empirical PDF (x points and their corrersponding PDF(x) values) and then sampled using
an interpolation of the inverse CDF function.
"""

from os import urandom

import numpy as np
from scipy.interpolate import interp1d

from tdastro.base_models import FunctionNode


class DistributionSampler(FunctionNode):
    """A node for sampling from an arbitrary one-dimensional distribution
    using inverse transformation sampling.

    Attributes
    ----------
    _rng : numpy.random._generator.Generator
        This object's random number generator.
    _inv_cdf : scipy.interpolate.interp1d
        The interpolation function for the inverse CDF. Maps values [0, 1]
        to the corresponding x values.

    Parameters
    ----------
    x : numpy.ndarray[float]
        The random variable (input to the density function)
    y : numpy.ndarray[float]
        The probability density function (output).
    kind : str
        The mode of interpolation for CDF function.
        Must be one of "linear", "nearest", "nearest-up", "zero", "slinear",
        "quadratic", "cubic", "previous", or "next".
        Default: "linear"
    """

    def __init__(self, x, y, kind="linear", seed=None, **kwargs):
        super().__init__(self._non_func, **kwargs)

        # Get a default random number generator for this object, using the
        # given seed if one is provided.
        if seed is None:
            seed = int.from_bytes(urandom(4), "big")
        self._rng = np.random.default_rng(seed=seed)

        # Validate the inputs.
        if len(x) != len(y):
            raise ValueError(f"Length of x ({len(x)}) and y ({len(y)}) must be equal.")
        if np.any(y < 0):
            raise ValueError("Probability density function (y) must be non-negative.")

        # Create the CDF values (and normalize it to 1) and then create an interpolation
        # of the inverse CDF function.
        cdf_samples = np.cumsum(y)
        cdf_samples /= cdf_samples[-1]
        self._inv_cdf = interp1d(cdf_samples, x, kind=kind, fill_value=0.0)

    def set_seed(self, new_seed):
        """Update the random number generator's seed to a given value.

        Parameters
        ----------
        new_seed : int
            The given seed
        """
        self._rng = np.random.default_rng(seed=new_seed)

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Draw samples from the inverse PDF.

        The input arguments are taken from the current graph_state and the outputs
        are written to graph_state.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        results : any
            The result of the computation. This return value is provided so that testing
            functions can easily access the results.
        """
        num_samples = None if graph_state.num_samples == 1 else graph_state.num_samples

        # Draw samples of the CDF [0, 1] and then use the inverse CDF spline to map those
        # back to the corresponding x values.
        cdf_points = self._rng.random(num_samples)
        x_samples = self._inv_cdf(cdf_points)
        self._save_results(x_samples, graph_state)
        return x_samples



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html


import os
import sys
from importlib.metadata import version

# Define path to the code to be documented **relative to where conf.py (this file) is kept**
sys.path.insert(0, os.path.abspath("../src/"))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "tdastro"
copyright = "2024, LINCC Frameworks"
author = "LINCC Frameworks"
release = version("tdastro")
# for example take major/minor
version = ".".join(release.split(".")[:2])

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.mathjax", "sphinx.ext.napoleon", "sphinx.ext.viewcode"]

extensions.append("autoapi.extension")
extensions.append("nbsphinx")

# -- sphinx-copybutton configuration ----------------------------------------
extensions.append("sphinx_copybutton")
## sets up the expected prompt text from console blocks, and excludes it from
## the text that goes into the clipboard.
copybutton_exclude = ".linenos, .gp"
copybutton_prompt_text = ">> "

## lets us suppress the copy button on select code blocks.
copybutton_selector = "div:not(.no-copybutton) > div.highlight > pre"

templates_path = []
exclude_patterns = ["_build", "**.ipynb_checkpoints"]

# This assumes that sphinx-build is called from the root directory
master_doc = "index"
# Remove 'view source code' from top of page (for html, not python)
html_show_sourcelink = False
# Remove namespaces from class/method signatures
add_module_names = False

autoapi_type = "python"
autoapi_dirs = ["../src"]
autoapi_ignore = ["*/__main__.py", "*/_version.py"]
autoapi_add_toc_tree_entry = False
autoapi_member_order = "bysource"

html_theme = "sphinx_rtd_theme"



"""A simple node used for testing."""

from tdastro.base_models import ParameterizedNode


class SingleVariableNode(ParameterizedNode):
    """A ParameterizedNode holding a single pre-defined variable.

    Notes
    -----
    Often used for testing, but can be used to make graph dependencies clearer.

    Parameters
    ----------
    name : str
        The parameter name.
    value : any
        The parameter value.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(self, name, value, **kwargs):
        super().__init__(**kwargs)
        self.add_parameter(name, value, **kwargs)



import warnings
from pathlib import Path
from unittest.mock import patch

import numpy as np
import pytest
from sncosmo import Bandpass
from tdastro.astro_utils.passbands import Passband
from tdastro.sources.spline_model import SplineModel


def create_lsst_passband(path, filter_name, **kwargs):
    """Helper function to create an LSST Passband object for testing."""
    survey = "LSST"
    table_path = f"{path}/{survey}/{filter_name}.dat"
    return Passband.from_file(survey, filter_name, table_path=table_path, **kwargs)


def create_toy_passband(path, transmission_table, filter_name="a", **kwargs):
    """Helper function to create a toy Passband object for testing."""
    survey = "TOY"
    dir_path = Path(path, survey)
    dir_path.mkdir(parents=True, exist_ok=True)
    table_path = dir_path / f"{filter_name}.dat"

    # Create a transmission table file
    with open(table_path, "w") as f:
        f.write(transmission_table)

    # Create a Passband object
    return Passband.from_file(survey, filter_name, table_path=table_path, **kwargs)


def test_passband_str(passbands_dir, tmp_path):
    """Test the __str__ method of the Passband class."""
    # Test an LSST passband:
    LSST_u = create_lsst_passband(passbands_dir, "u")
    assert str(LSST_u) == "Passband: LSST_u"

    # Test a toy passband:
    transmission_table = "1000 0.5\n1005 0.6\n1010 0.7\n"
    a_band = create_toy_passband(tmp_path, transmission_table)
    assert str(a_band) == "Passband: TOY_a"


def test_passband_eq(passbands_dir, tmp_path):
    """Test the __str__ method of the Passband class."""
    a_band = Passband(np.array([[1000, 0.5], [1005, 0.6], [1010, 0.7]]), "LSST", "a")
    b_band = Passband(np.array([[1000, 0.5], [1005, 0.6], [1010, 0.7]]), "LSST", "b")
    c_band = Passband(np.array([[1000, 0.5], [1005, 0.7], [1010, 0.7]]), "LSST", "c")
    d_band = Passband(np.array([[1000, 0.5], [1005, 0.6], [1020, 0.7]]), "LSST", "d")
    assert a_band == b_band
    assert a_band != c_band
    assert a_band != d_band


def test_passband_manual_create(tmp_path):
    """Test that we can create a passband from the transmission table."""
    # Test we get a TypeError if we don't provide a survey and filter_name
    with pytest.raises(TypeError):
        _ = Passband()

    # Create a manual passband.
    transmission_table = np.array([[1000, 0.5], [1005, 0.6], [1010, 0.7]])
    test_pb = Passband(transmission_table, survey="test", filter_name="u")

    assert test_pb.survey == "test"
    assert test_pb.filter_name == "u"
    assert test_pb.full_name == "test_u"
    np.testing.assert_allclose(test_pb._loaded_table, transmission_table)

    # We can create an load a table in nm as well. It will auto-convert to Angstroms.
    transmission2 = np.array([[100, 0.5], [100.5, 0.6], [101, 0.7]])
    test_pb2 = Passband(transmission2, survey="test", filter_name="g", units="nm")
    assert test_pb2.survey == "test"
    assert test_pb2.filter_name == "g"
    assert test_pb2.full_name == "test_g"
    np.testing.assert_allclose(test_pb2._loaded_table, transmission_table)

    # We raise an error if the data is not sorted.
    with pytest.raises(ValueError):
        _ = Passband(
            np.array([[1000, 0.5], [900, 0.2], [1100, 0.7]]),
            survey="test",
            filter_name="u",
        )

    # We raise an error if the data is not the right shape.
    with pytest.raises(ValueError):
        _ = Passband(
            np.full((10, 3), 1.0),
            survey="test",
            filter_name="u",
        )

    # We raise an error with an invalid unit.
    with pytest.raises(ValueError):
        _ = Passband(
            transmission_table,
            survey="test",
            filter_name="u",
            units="invalid",
        )


def test_passband_load_transmission_table(passbands_dir, tmp_path):
    """Test the _load_transmission_table method of the Passband class."""
    # Test that we can load a standard LSST transmission table from the local test data
    LSST_g = create_lsst_passband(passbands_dir, "g")
    assert LSST_g._loaded_table is not None
    assert LSST_g._loaded_table.shape[1] == 2  # Two columns: wavelengths and transmissions
    assert LSST_g.waves is not None
    assert len(LSST_g.waves.shape) == 1

    # Test a toy transmission table was loaded correctly
    transmission_table = "1000 0.5\n1005 0.6\n1010 0.7\n"
    a_band = create_toy_passband(tmp_path, transmission_table)
    np.testing.assert_allclose(a_band._loaded_table, np.array([[1000, 0.5], [1005, 0.6], [1010, 0.7]]))

    # Test that we raise an error if the transmission table is blank
    transmission_table = ""
    test_pb_file_name = Path(tmp_path) / "r.dat"
    with open(test_pb_file_name, "w") as f:
        f.write(transmission_table)
    with warnings.catch_warnings():
        # Ignore the warning that we are loading an empty file.
        warnings.simplefilter("ignore")
        with pytest.raises(ValueError):
            a_band.load_transmission_table(test_pb_file_name)

    # Test that we raise an error if the transmission table is not formatted correctly
    transmission_table = "1000\n1005 0.6\n1010 0.7"
    with open(test_pb_file_name, "w") as f:
        f.write(transmission_table)
    with pytest.raises(ValueError):
        a_band.load_transmission_table(test_pb_file_name)

    # Test that we raise an error if the transmission table wavelengths are not sorted
    transmission_table = "1000 0.5\n900 0.6\n1010 0.7"
    with open(test_pb_file_name, "w") as f:
        f.write(transmission_table)
    with pytest.raises(ValueError):
        a_band.load_transmission_table(test_pb_file_name)


def test_passband_download_transmission_table(tmp_path):
    """Test the functionality of downloading the transmission table Passband class."""
    # Initialize a Passband object
    survey = "TEST"
    filter_name = "a"
    table_path = tmp_path / f"{survey}/{filter_name}.dat"
    table_url = f"https://data.lsdb.io/{survey}/{filter_name}.dat"
    transmission_table = "1000 0.5\n1005 0.6\n1010 0.7\n"

    def mock_urlretrieve(url, known_hash, fname, path):
        full_name = path / fname
        with open(full_name, "w") as f:
            f.write(transmission_table)
        return full_name

    # Mock the urlretrieve portion of the download method
    with patch("pooch.retrieve", side_effect=mock_urlretrieve) as mocked_urlretrieve:
        a_band = Passband.from_file(survey, filter_name, table_path=table_path, table_url=table_url)

        # Check that the transmission table was downloaded
        mocked_urlretrieve.assert_called_once_with(
            url=table_url,
            known_hash=None,
            fname=table_path.name,
            path=table_path.parent,
        )

        # Check that the transmission table was loaded correctly
        np.testing.assert_allclose(a_band._loaded_table, np.array([[1000, 0.5], [1005, 0.6], [1010, 0.7]]))


def test_passband_from_file(passbands_dir, tmp_path):
    """Test the from_file constructor of the Passband class."""
    # Test a toy transmission table was loaded correctly
    transmission_table = "1000 0.5\n1005 0.6\n1010 0.7\n"
    a_band = create_toy_passband(tmp_path, transmission_table)
    np.testing.assert_allclose(a_band._loaded_table, np.array([[1000, 0.5], [1005, 0.6], [1010, 0.7]]))

    table_path = Path(tmp_path, "TOY", "a.dat")
    a_band_2 = Passband.from_file("TOY", "a", table_path=table_path)
    np.testing.assert_allclose(a_band_2._loaded_table, np.array([[1000, 0.5], [1005, 0.6], [1010, 0.7]]))

    # We raise an error if the transmission table does not exist.
    with pytest.raises(ValueError):
        _ = Passband.from_file(survey="test", filter_name="u", table_path="./no_such_file.dat")


def test_passband_from_sncosmo(passbands_dir):
    """Test the from_sncosmo constructor of the Passband class."""
    sn_pb = Bandpass(
        np.array([6000, 6005, 6010]),  # wavelengths (A)
        np.array([0.5, 0.6, 0.7]),  # transmissions
    )
    ztf_band = Passband.from_sncosmo("ZTF", "g", sn_pb)
    assert ztf_band.survey == "ZTF"
    assert ztf_band.filter_name == "g"
    assert ztf_band.full_name == "ZTF_g"
    assert ztf_band._loaded_table is not None
    assert np.allclose(ztf_band._loaded_table[:, 0], sn_pb.wave)
    assert np.allclose(ztf_band._loaded_table[:, 1], sn_pb.trans)


def test_process_transmission_table(passbands_dir, tmp_path):
    """Test the process_transmission_table method of the Passband class; check correct methods are called."""
    transmission_table = "100 0.5\n200 0.75\n300 0.25\n"
    a_band = create_toy_passband(tmp_path, transmission_table)

    # Mock the methods that _process_transmission_table wraps, to check each is called once
    with (
        patch.object(a_band, "_interpolate_transmission_table") as mock_interp_table,
        patch.object(a_band, "_trim_transmission_by_quantile") as mock_trim_table,
        patch.object(a_band, "_normalize_transmission_table") as mock_norm_table,
    ):
        # Call the _process_transmission_table method
        delta_wave, trim_quantile = 1.0, 0.05
        a_band.process_transmission_table(delta_wave, trim_quantile)

        # Check that each method is called once
        mock_interp_table.assert_called_once_with(a_band._loaded_table, delta_wave)
        mock_trim_table.assert_called_once()
        mock_norm_table.assert_called_once()

    # Now call without mocking, to check waves set correctly (other values checked in method-specific tests)
    delta_wave, trim_quantile = 5.0, None
    a_band.process_transmission_table(delta_wave, trim_quantile)
    np.testing.assert_allclose(a_band.waves, np.arange(100, 301, delta_wave))

    # Check that we can call the method on a standard LSST transmission table
    LSST_r = create_lsst_passband(passbands_dir, "r")
    LSST_r.process_transmission_table(delta_wave, trim_quantile)
    assert LSST_r.waves is not None

    # Check that we raise an error if the transmission table is the wrong shape
    a_band._loaded_table = np.array([[100, 0.5, 105, 0.6]])
    with pytest.raises(ValueError):
        a_band.process_transmission_table(delta_wave, trim_quantile)


def test_interpolate_transmission_table(passbands_dir, tmp_path):
    """Test the _interpolate_transmission_table method of the Passband class."""
    transmission_table = "100 0.5\n200 0.75\n300 0.25\n"
    a_band = create_toy_passband(tmp_path, transmission_table)

    # Interpolate the transmission table to a large step size (50 Angstrom)
    table = a_band._interpolate_transmission_table(a_band._loaded_table, delta_wave=50)
    np.testing.assert_allclose(table[:, 0], np.arange(100, 301, 50))

    # Interpolate the transmission table to a somewhat small step size (1 Angstrom)
    table = a_band._interpolate_transmission_table(a_band._loaded_table, delta_wave=1)
    np.testing.assert_allclose(table[:, 0], np.arange(100, 301, 1))

    # Interpolate the transmission table to an even smaller step size (0.1 Angstrom)
    table = a_band._interpolate_transmission_table(a_band._loaded_table, delta_wave=0.1)
    np.testing.assert_allclose(table[:, 0], np.arange(100, 300.01, 0.1))

    # Check that we raise an error if the transmission table is the wrong shape
    with pytest.raises(ValueError):
        a_band._interpolate_transmission_table(np.array([[100, 0.5], [200, 0.75], [300, 0.25], [400]]))

    # Check that we raise an error if the delta_wave is not a positive number
    with pytest.raises(ValueError):
        a_band._interpolate_transmission_table(a_band._loaded_table, delta_wave=-1)

    # Check that we can run method on a standard LSST transmission table
    LSST_i = create_lsst_passband(passbands_dir, "i")
    table = LSST_i._interpolate_transmission_table(LSST_i._loaded_table, delta_wave=50)
    assert len(table) > 0
    assert len(table) < len(LSST_i._loaded_table)


def test_trim_transmission_table(passbands_dir, tmp_path):
    """Test the _trim_transmission_by_quantile method of the Passband class."""

    # Test: transmission table with only 3 points
    transmission_table = "100 0.5\n200 0.75\n300 0.25\n"
    a_band = create_toy_passband(tmp_path, transmission_table)

    # Trim the transmission table by 5% (should remove the last point)
    table = a_band._trim_transmission_by_quantile(a_band._loaded_table, trim_quantile=0.05)
    np.testing.assert_allclose(table, np.array([[100, 0.5], [200, 0.75]]))

    # Trim the transmission table by 40% (should still preserve the first point, as table is skewed)
    table = a_band._trim_transmission_by_quantile(a_band._loaded_table, trim_quantile=0.4)
    np.testing.assert_allclose(table, np.array([[100, 0.5], [200, 0.75]]))

    # Check no trimming if quantile is 0
    table = a_band._trim_transmission_by_quantile(a_band._loaded_table, trim_quantile=None)
    np.testing.assert_allclose(table, np.array([[100, 0.5], [200, 0.75], [300, 0.25]]))

    # Check we raise an error if the quantile is not greater or equal to 0 and less than 0.5
    with pytest.raises(ValueError):
        a_band._trim_transmission_by_quantile(a_band._loaded_table, trim_quantile=0.5)
    with pytest.raises(ValueError):
        a_band._trim_transmission_by_quantile(a_band._loaded_table, trim_quantile=-0.1)

    # Check we raise an error if the transmission table is the wrong shape
    with pytest.raises(ValueError):
        a_band._trim_transmission_by_quantile(np.array([[100, 0.5], [200, 0.75], [300, 0.25], [400]]))

    # Test 2: larger, more gaussian transmission table
    transmission_table = "100 0.05\n200 0.1\n300 0.25\n400 0.5\n500 0.8\n600 0.6\n700 0.4\n800 0.2\n900 0.1\n"
    b_band = create_toy_passband(tmp_path, transmission_table, filter_name="b")

    # Trim the transmission table by 5% (should remove the last point)
    table = b_band._trim_transmission_by_quantile(b_band._loaded_table, trim_quantile=0.05)
    np.testing.assert_allclose(
        table,
        np.array(
            [[100, 0.05], [200, 0.1], [300, 0.25], [400, 0.5], [500, 0.8], [600, 0.6], [700, 0.4], [800, 0.2]]
        ),
    )

    # Trim the transmission table by 40% (should remove most points)
    table = b_band._trim_transmission_by_quantile(b_band._loaded_table, trim_quantile=0.4)
    np.testing.assert_allclose(table, np.array([[300, 0.25], [400, 0.5], [500, 0.8]]))

    # Check no trimming if quantile is 0
    table = b_band._trim_transmission_by_quantile(b_band._loaded_table, trim_quantile=None)
    np.testing.assert_allclose(
        table,
        np.array(
            [
                [100, 0.05],
                [200, 0.1],
                [300, 0.25],
                [400, 0.5],
                [500, 0.8],
                [600, 0.6],
                [700, 0.4],
                [800, 0.2],
                [900, 0.1],
            ]
        ),
    )

    # Test 3: much larger transmission table
    rng = np.random.default_rng(100)
    transmissions = rng.normal(0.5, 0.1, 1000)
    wavelengths = np.arange(100, 1100, 1)
    transmission_table = "\n".join(
        [
            f"{wavelength} {transmission}"
            for wavelength, transmission in zip(wavelengths, transmissions, strict=False)
        ]
    )
    c_band = create_toy_passband(tmp_path, transmission_table, filter_name="c")

    # Trim the transmission table by 5% on each side
    table = c_band._trim_transmission_by_quantile(c_band._loaded_table, trim_quantile=0.05)
    assert len(table) < len(c_band._loaded_table)

    original_area = np.trapezoid(c_band._loaded_table[:, 1], x=c_band._loaded_table[:, 0])
    trimmed_area = np.trapezoid(table[:, 1], x=table[:, 0])
    assert trimmed_area >= (original_area * 0.9)

    # Test 4: LSST transmission table
    LSST_z = create_lsst_passband(passbands_dir, "z")
    table = LSST_z._trim_transmission_by_quantile(LSST_z._loaded_table, trim_quantile=0.05)
    assert len(table) < len(LSST_z._loaded_table)

    original_area = np.trapezoid(LSST_z._loaded_table[:, 1], x=LSST_z._loaded_table[:, 0])
    trimmed_area = np.trapezoid(table[:, 1], x=table[:, 0])
    assert trimmed_area >= (original_area * 0.9)


def test_passband_normalize_transmission_table(passbands_dir, tmp_path):
    """Test the _normalize_transmission_table method of the Passband class."""
    # Test a toy transmission table
    transmission_table = "100 0.5\n200 0.75\n300 0.25\n"
    a_band = create_toy_passband(tmp_path, transmission_table, delta_wave=100, trim_quantile=None)

    # Normalize the transmission table (we skip interpolation as grid already matches)
    a_band._normalize_transmission_table(a_band._loaded_table)

    # Compare results
    expected_result = np.array([[100.0, 0.0075], [200.0, 0.005625], [300.0, 0.00125]])
    assert np.allclose(a_band.processed_transmission_table, expected_result)

    # Test that we raise an error if the transmission table is the wrong size or shape
    with pytest.raises(ValueError):
        a_band._normalize_transmission_table(np.array([[]]))
    with pytest.raises(ValueError):
        a_band._normalize_transmission_table(np.array([[100, 0.5]]))
    with pytest.raises(ValueError):
        a_band._normalize_transmission_table(np.array([100, 0.5]))
    with pytest.raises(ValueError):
        a_band._normalize_transmission_table(np.array([[100, 0.5, 105, 0.6]]))

    # Test we can call method on a standard LSST transmission table
    LSST_y = create_lsst_passband(passbands_dir, "y")
    normalized_table = LSST_y._normalize_transmission_table(LSST_y._loaded_table)
    assert len(normalized_table) == len(LSST_y._loaded_table)
    assert not np.allclose(normalized_table, LSST_y._loaded_table)


def test_passband_fluxes_to_bandflux(passbands_dir, tmp_path):
    """Test the fluxes_to_bandflux method of the Passband class."""
    transmission_table = "100 0.5\n200 0.75\n300 0.25\n"
    a_band = create_toy_passband(tmp_path, transmission_table, delta_wave=100, trim_quantile=None)

    # Define some mock flux values and calculate our expected bandflux
    flux = np.array(
        [
            [1.0, 1.0, 1.0],
            [2.0, 1.0, 1.0],
            [3.0, 1.0, 1.0],
            [2.0, 1.0, 1.0],
            [1.0, 1.0, 1.0],
        ]
    )
    expected_in_band_flux = np.trapezoid(flux * a_band.processed_transmission_table[:, 1], x=a_band.waves)
    in_band_flux = a_band.fluxes_to_bandflux(flux)
    np.testing.assert_allclose(in_band_flux, expected_in_band_flux)

    # Test with a different set of fluxes, regridding the transmission table
    a_band.process_transmission_table(delta_wave=50, trim_quantile=None)
    flux = np.array(
        [
            [100.0, 12.0, 1.0, 0.5, 0.25],
            [300.0, 200.0, 100.0, 50.0, 25.0],
            [1.0, 500, 1000, 500, 250],
            [2.0, 1.0, 1.0, 1.0, 1.0],
            [3.0, 1.0, 1.0, 1.0, 1.0],
            [200.0, 100.0, 50.0, 25.0, 12.5],
            [100.0, 50.0, 25.0, 12.5, 6.25],
        ]
    )
    in_band_flux = a_band.fluxes_to_bandflux(flux)
    expected_in_band_flux = np.trapezoid(
        flux * a_band.processed_transmission_table[:, 1], x=a_band.processed_transmission_table[:, 0]
    )
    np.testing.assert_allclose(in_band_flux, expected_in_band_flux)

    # Test we raise an error if the fluxes are not the right shape
    with pytest.raises(ValueError):
        a_band.fluxes_to_bandflux(np.array([[1.0, 2.0], [3.0, 4.0]]))
    with pytest.raises(ValueError):
        a_band.fluxes_to_bandflux(np.full((7, 5, 2, 2), 1.0))

    # Test we raise an error if the fluxes are empty
    with pytest.raises(ValueError):
        a_band.fluxes_to_bandflux(np.array([]))

    # Test we can call method on a standard LSST transmission table
    LSST_u = create_lsst_passband(passbands_dir, "u")
    flux = np.random.rand(5, len(LSST_u.waves))
    in_band_flux = LSST_u.fluxes_to_bandflux(flux)
    assert in_band_flux is not None
    assert len(in_band_flux) == 5


def test_passband_fluxes_to_bandflux_mult_samples(passbands_dir, tmp_path):
    """Test the fluxes_to_bandflux method of the Passband class with multiple samples."""
    transmission_table = "100 0.5\n200 0.75\n300 0.25\n"
    a_band = create_toy_passband(tmp_path, transmission_table, delta_wave=100, trim_quantile=None)

    # Define some mock flux values and calculate our expected bandflux
    flux = np.array(
        [
            [
                [1.0, 1.0, 1.0],
                [2.0, 1.0, 1.0],
                [3.0, 1.0, 1.0],
                [2.0, 1.0, 1.0],
                [1.0, 1.0, 1.0],
            ],
            [
                [2.0, 2.0, 2.0],
                [4.0, 2.0, 2.0],
                [6.0, 2.0, 2.0],
                [4.0, 2.0, 2.0],
                [2.0, 2.0, 2.0],
            ],
        ]
    )
    expected = np.array([[1.0, 1.375, 1.75, 1.375, 1.0], [2.0, 2.75, 3.5, 2.75, 2.0]])

    result = a_band.fluxes_to_bandflux(flux)
    np.testing.assert_allclose(result, expected)


def test_passband_wrapped_from_physical_source(passbands_dir, tmp_path):
    """Test get_band_fluxes, PhysicalModel's wrapped version of Passband's fluxes_to_bandflux.."""
    # Set up physical model
    times = np.array([1.0, 2.0, 3.0])
    wavelengths = np.array([10.0, 20.0, 30.0])
    fluxes = np.array([[1.0, 5.0, 1.0], [5.0, 10.0, 5.0], [1.0, 5.0, 3.0]])
    model = SplineModel(times, wavelengths, fluxes, time_degree=1, wave_degree=1)
    state = model.sample_parameters()

    test_times = np.array([1.0, 1.5, 2.0, 2.5, 3.0, 3.5])

    # Test with a single toy passband (see PassbandGroup tests for group tests)
    transmission_table = "100 0.5\n200 0.75\n300 0.25\n"
    a_band = create_toy_passband(tmp_path, transmission_table, delta_wave=100, trim_quantile=None)
    result_from_source_model = model.get_band_fluxes(a_band, test_times, filters=None, state=state)

    evaluated_fluxes = model.evaluate(test_times, a_band.waves, state)
    result_from_passband = a_band.fluxes_to_bandflux(evaluated_fluxes)
    np.testing.assert_allclose(result_from_source_model, result_from_passband)

    # Test with a standard LSST passband
    LSST_g = create_lsst_passband(passbands_dir, "g")
    result_from_source_model = model.get_band_fluxes(
        LSST_g, test_times, filters=np.repeat("g", len(test_times)), state=state
    )

    evaluated_fluxes = model.evaluate(test_times, LSST_g.waves, state)
    result_from_passband = LSST_g.fluxes_to_bandflux(evaluated_fluxes)
    np.testing.assert_allclose(result_from_source_model, result_from_passband)



#!/usr/bin/env python

import argparse


def parse_args(args):
    """Parse the command line arguments.

    Parameters
    ----------
    args : `list` of `str`
        The command line arguments.

    Returns
    -------
    `argparse.Namespace`
        The parsed arguments.
    """
    parser = argparse.ArgumentParser(description="Create a truncated version of a SALT2 table.")
    parser.add_argument("--input", default="", help="The input SALT2 data file.")
    parser.add_argument("--output", default="", help="The file name for the output data.")
    parser.add_argument("--t_min", default=-100.0, help="The minimum time to include.")
    parser.add_argument("--t_max", default=100.0, help="The maxmimum time to include.")
    parser.add_argument("--w_min", default=0.0, help="The minimum wavelength to include.")
    parser.add_argument("--w_max", default=25000.0, help="The maxmimum wavelength to include.")
    return parser.parse_args(args)


def main(args=None):
    """Subsample an OpSim file and save it to a new file.

    Parameters
    ----------
    args : `list` of `str`, optional
        The command line arguments.
    """
    args = parse_args(args)
    if args.input == args.output:
        raise ValueError("Cannot overwrite file.")
    times = (float(args.t_min), float(args.t_max))
    waves = (float(args.w_min), float(args.w_max))

    print(f"Input file: {args.input}")
    print(f"Output file: {args.output}")
    print(f"Time range: {times}")
    print(f"Wavelength range: {waves}")

    with open(args.input, "r") as input_file:
        with open(args.output, "w") as output_file:
            for line in input_file:
                tokens = line.split()
                if (times[0] <= float(tokens[0]) <= times[1]) and (waves[0] <= float(tokens[1]) <= waves[1]):
                    output_file.write(line)


if __name__ == "__main__":
    main()



import numpy as np
from tdastro.math_nodes.ra_dec_sampler import (
    OpSimRADECSampler,
    OpSimUniformRADECSampler,
    UniformRADEC,
)
from tdastro.opsim.opsim import OpSim


def test_uniform_ra_dec():
    """Test that we can generate numbers from a uniform distribution on a sphere."""
    sampler_node = UniformRADEC(seed=100, node_label="sampler")

    # Test we can generate a single value.
    (ra, dec) = sampler_node.generate(num_samples=1)
    assert 0.0 <= ra <= 360.0
    assert -90.0 <= dec <= 90.0

    # Generate many samples.
    num_samples = 20_000
    state = sampler_node.sample_parameters(num_samples=num_samples)

    all_ra = state["sampler"]["ra"]
    assert len(all_ra) == num_samples
    assert np.all(all_ra >= 0.0)
    assert np.all(all_ra <= 360.0)

    all_dec = state["sampler"]["dec"]
    assert len(all_dec) == num_samples
    assert np.all(all_dec >= -90.0)
    assert np.all(all_dec <= 90.0)

    # Compute histograms of RA and dec values.
    ra_bins = np.zeros(36)
    dec_bins = np.zeros(18)
    for idx in range(num_samples):
        ra_bins[int(all_ra[idx] / 10.0)] += 1
        dec_bins[int((all_dec[idx] + 90.0) / 10.0)] += 1

    # Check that all RA bins have approximately equal samples.
    expected_count = num_samples / 36
    for bin_count in ra_bins:
        assert 0.8 <= bin_count / expected_count <= 1.2

    # Check that the dec bins around the poles have less samples
    # than the bins around the equator.
    assert dec_bins[0] < 0.25 * dec_bins[9]
    assert dec_bins[17] < 0.25 * dec_bins[10]

    # Check that we can generate uniform samples in radians.
    sampler_node2 = UniformRADEC(seed=100, node_label="sampler2", use_degrees=False)
    state2 = sampler_node2.sample_parameters(num_samples=num_samples)

    all_ra = state2["sampler2"]["ra"]
    assert len(all_ra) == num_samples
    assert np.all(all_ra >= 0.0)
    assert np.all(all_ra <= 2.0 * np.pi)

    all_dec = state2["sampler2"]["dec"]
    assert len(all_dec) == num_samples
    assert np.all(all_dec >= -np.pi)
    assert np.all(all_dec <= np.pi)


def test_opsim_ra_dec_sampler():
    """Test that we can sample from am OpSim object."""
    values = {
        "observationStartMJD": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "fieldRA": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "fieldDec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "zp_nJy": np.ones(5),
    }
    ops_data = OpSim(values)
    assert len(ops_data) == 5

    sampler_node = OpSimRADECSampler(ops_data, in_order=True)

    # Test we can generate a single value.
    (ra, dec, time) = sampler_node.generate(num_samples=1)
    assert ra == 15.0
    assert dec == -10.0
    assert time == 0.0

    # Test we can generate multiple observations
    (ra, dec, time) = sampler_node.generate(num_samples=2)
    assert np.allclose(ra, [30.0, 15.0])
    assert np.allclose(dec, [-5.0, 0.0])
    assert np.allclose(time, [1.0, 2.0])

    # Do randomized sampling.
    sampler_node2 = OpSimRADECSampler(ops_data, in_order=False, seed=100, node_label="sampler")
    state = sampler_node2.sample_parameters(num_samples=5000)

    # Check that the samples are uniform and consistent.
    int_times = state["sampler"]["time"].astype(int)
    assert np.allclose(state["sampler"]["ra"], values["fieldRA"][int_times])
    assert np.allclose(state["sampler"]["dec"], values["fieldDec"][int_times])
    assert len(int_times[int_times == 0]) > 750
    assert len(int_times[int_times == 1]) > 750
    assert len(int_times[int_times == 2]) > 750
    assert len(int_times[int_times == 3]) > 750
    assert len(int_times[int_times == 4]) > 750

    # Do randomized sampling with offsets.
    sampler_node3 = OpSimRADECSampler(ops_data, in_order=False, seed=100, radius=0.1, node_label="sampler")
    state = sampler_node3.sample_parameters(num_samples=5000)

    # Check that the samples are not all the centers (unique values > 5) but are close.
    int_times = state["sampler"]["time"].astype(int)
    assert len(np.unique(state["sampler"]["ra"])) > 5
    assert len(np.unique(state["sampler"]["dec"])) > 5
    assert np.allclose(state["sampler"]["ra"], values["fieldRA"][int_times], atol=0.2)
    assert np.allclose(state["sampler"]["dec"], values["fieldDec"][int_times], atol=0.2)


def test_opsim_uniform_ra_dec_sampler():
    """Test that we can sample uniformly from am OpSim object."""
    # Create an opsim with two points in different hemispheres.
    values = {
        "observationStartMJD": np.array([0.0, 1.0]),
        "fieldRA": np.array([15.0, 195.0]),
        "fieldDec": np.array([75.0, -75.0]),
        "zp_nJy": np.ones(2),
    }
    ops_data = OpSim(values)
    assert len(ops_data) == 2

    # Use a very large radius so we do not reject too many samples.
    sampler_node = OpSimUniformRADECSampler(ops_data, radius=70.0, seed=100, node_label="sampler")

    # Test we can generate a single value.
    ra, dec = sampler_node.generate(num_samples=1)
    assert ops_data.is_observed(ra, dec, radius=70.0)

    # Test we can generate many observations
    num_samples = 10_000
    ra, dec = sampler_node.generate(num_samples=num_samples)
    assert np.all(ops_data.is_observed(ra, dec, radius=70.0))

    # We should sample roughly uniformly from the two regions.
    northern_mask = dec > 0.0
    assert np.sum(northern_mask) > 0.4 * num_samples
    assert np.sum(northern_mask) < 0.6 * num_samples



"""The core functions for running the TDAstro simulation."""

import numpy as np
import pandas as pd
from nested_pandas import NestedFrame

from tdastro.astro_utils.noise_model import apply_noise


def simulate_lightcurves(
    source,
    num_samples,
    opsim,
    passbands,
    opsim_save_cols=None,
    param_cols=None,
    apply_obs_mask=False,
    rng=None,
):
    """Generate a number of simulations of the given source.

    Parameters
    ----------
    source : PhysicalModel
        The source to draw from. This may have its own parameters which
        will be randomly sampled with each draw.
    num_samples : int
        The number of samples.
    opsim : OpSim
        The OpSim information for the samples.
    passbands : PassbandGroup
        The passbands to use for generating the bandfluxes.
    apply_obs_mask: boolean
        If True, apply obs_mask to filter interesting indices/times
    opsim_save_cols : list of str, optional
        A list of opsim columns to be saved as part of the results. This is used
        to save context information about how the lightcurves were generated.
        If None, no additional columns are saved.
    param_cols : list of str, optional
        A list of source parameter columns to be saved as separate columns in
        the results (instead of just the full dictionary of parameters). These
        must be specified as strings in the node_name.param_name format.
        If None, no additional columns are saved.
    rng : numpy.random._generator.Generator, optional
        A given numpy random number generator to use for this computation. If not
        provided, the function uses the node's random number generator.

    Returns
    -------
    lightcurves : nested_pandas.NestedFrame
        A NestedFrame with a row for each object.
    """
    # Sample the parameter space of this model.
    if num_samples <= 0:
        raise ValueError("Invalid number of samples.")
    sample_states = source.sample_parameters(num_samples=num_samples, rng_info=rng)

    # Determine which of the of the simulated positions match opsim locations.
    ra = source.get_param(sample_states, "ra")
    dec = source.get_param(sample_states, "dec")
    all_obs_matches = opsim.range_search(ra, dec)

    # Get all times and all filters as numpy arrays so we can do easy subsets.
    all_times = np.asarray(opsim["time"].values, dtype=float)
    all_filters = np.asarray(opsim["filter"].values, dtype=str)

    # Create dictionaries for keeping all the result information. The first
    # stores per-object information and the second per-object, per-observation.
    # nested_index maps the entry in the nested array to the object's index.
    results_dict = {
        "id": [],
        "ra": [],
        "dec": [],
        "nobs": [],
        "z": [],
        "params": [],
    }
    nested_dict = {
        "mjd": [],
        "filter": [],
        "flux": [],
        "fluxerr": [],
        "flux_perfect": [],
    }
    nested_index = []

    # Add the extra columns to both the results and nested dictionaries.
    if opsim_save_cols is None:
        opsim_save_cols = []
    for col in opsim_save_cols:
        nested_dict[col] = []
    if param_cols is None:
        param_cols = []
    for col in param_cols:
        results_dict[col.replace(".", "_")] = []

    for idx, state in enumerate(sample_states):
        # Find the indices and times where the current source is seen.
        obs_index = np.asarray(all_obs_matches[idx])
        if len(obs_index) == 0:
            obs_times = []
            obs_filters = []
        else:
            obs_times = all_times[obs_index]

            # Filter to only the "interesting" indices / times for this object.
            if apply_obs_mask:
                obs_mask = source.mask_by_time(obs_times, state)
                obs_index = obs_index[obs_mask]
                obs_times = obs_times[obs_mask]
            # Extract the filters for this observation.
            obs_filters = all_filters[obs_index]

        # Compute the band_fluxes and errors over just the given filters.
        bandfluxes_perfect = source.get_band_fluxes(passbands, obs_times, obs_filters, state)
        bandfluxes_error = opsim.bandflux_error_point_source(bandfluxes_perfect, obs_index)
        bandfluxes = apply_noise(bandfluxes_perfect, bandfluxes_error, rng=rng)

        # Save the object level information.
        results_dict["id"].append(idx)
        results_dict["ra"].append(ra[idx])
        results_dict["dec"].append(dec[idx])
        results_dict["nobs"].append(len(obs_times))
        results_dict["z"].append(source.get_param(state, "redshift"))
        results_dict["params"].append(state.to_dict())

        # Save the per-object parameters as separate columns. We use {node_name}_{param_name}
        # as the column name for each parameter since the . notation is used for nested columns.
        for col in param_cols:
            results_dict[col.replace(".", "_")].append(state[col])

        # Append the per-observation data to the nested dictionary, including
        # and needed opsim columns.
        nested_dict["mjd"].extend(list(obs_times))
        nested_dict["filter"].extend(list(obs_filters))
        nested_dict["flux_perfect"].extend(list(bandfluxes_perfect))
        nested_dict["flux"].extend(list(bandfluxes))
        nested_dict["fluxerr"].extend(list(bandfluxes_error))
        for col in opsim_save_cols:
            nested_dict[col].extend(list(opsim[col].values[obs_index]))

        nested_index.extend([idx] * len(obs_times))

    # Create the nested frame.
    results = NestedFrame(data=results_dict, index=[i for i in range(num_samples)])
    nested_frame = pd.DataFrame(data=nested_dict, index=nested_index)
    results = results.add_nested(nested_frame, "lightcurve")
    return results



import logging
from pathlib import Path
from typing import Literal, Union

import matplotlib.pyplot as plt
import numpy as np
import scipy.integrate
from citation_compass import cite_function
from sncosmo import Bandpass, get_bandpass

from tdastro import _TDASTRO_BASE_DATA_DIR
from tdastro.consts import lsst_filter_plot_colors
from tdastro.utils.data_download import download_data_file_if_needed

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


class PassbandGroup:
    """A group of passbands.

    The given passbands can come from a single survey or multiple surveys. As such, a given filter may appear
    in multiple passbands in a passband group. For example we might generate data from a combination of Rubin
    and DECCAM and want to use both their r filters. Thus the primary mapping is done by the passband's full
    name “{SURVEY}_{FILTER}”. Lookups by filter name are permitted in cases where the filter only occurs once
    in the group and thus the desired passband is unambiguous.

    Attributes
    ----------
    passbands : dict of Passband
        A dictionary of Passband objects, where the keys are the full_names of the passbands (eg, "LSST_u").
    table_dir : str
        The path to the directory containing the passband tables.
    waves : np.ndarray
        The union of all wavelengths in the passbands.
    _in_band_wave_indices : dict
        A dictionary mapping the passband name (eg, "LSST_u") to the indices of that specific
        passband's wavelengths in the full waves list.
    _filter_to_name : dict
        A dictionary mapping the filter name to a list of matching passband full names.
    """

    def __init__(self, given_passbands, filters=None, **kwargs):
        """Construct a PassbandGroup object.

        Parameters
        ----------
        given_passbands : Passband, list of Passband, or dict
            A list of all the passbands to include in this group. These can either be
            Passband objects or dictionaries of passband parameters with the following keys:
            - required:
                - survey : str - The name of the survey to which the passband belongs, e.g., "LSST".
                - filter_name : str - The name of the filter, e.g., "u".
            - one of the following:
                - table_data : np.ndarray - The transmission table data as a (N, 2) array where
                    the first column is wavelengths and the second column is transmission values.
                - table_path : str or Path - The path of the file from which to load the passband.
                - table_url : str - The URL from which to download the passband file.
            - optional:
                - delta_wave : float
                - trim_quantile : float
                - units : str (either 'nm' or 'A')
        filters : list, optional
            A list of filters to include in this PassbandGroup. If None, includes all filters.
            Otherwise drops filters that do not occur and throws an error if a filter is missing.
            Used for loading a subset of the filters.
        **kwargs
            Additional keyword arguments to pass to the Passband constructor.
        """
        self.passbands = {}
        self._in_band_wave_indices = {}
        self._filter_to_name = {}

        # If we are given a single Passband object or a single dictionary, wrap it in a list.
        if isinstance(given_passbands, Passband):
            given_passbands = [given_passbands]
        elif isinstance(given_passbands, dict):
            given_passbands = [given_passbands]
        elif given_passbands is None or len(given_passbands) == 0:
            raise ValueError("No passbands provided to PassbandGroup.")

        for pb in given_passbands:
            if isinstance(pb, Passband):
                # If we are given a Passband object, add it directly.
                self.passbands[pb.full_name] = pb
            elif isinstance(pb, dict):
                # If we are given a dictionary of parameters, create a Passband from it.
                params = pb.copy()

                # Add any missing parameters from kwargs.
                for key, value in kwargs.items():
                    if key not in params:
                        params[key] = value

                passband = Passband.from_file(**params)
                self.passbands[passband.full_name] = passband
            else:
                raise TypeError(f"Expected a Passband object or a dictionary of parameters. Got {type(pb)}.")

        # Prune any filters that are not on the given list and check for any missing filters.
        # We match on either the full name or the filter name.
        if filters is not None:
            filters = set(filters)
            filters_remaining = filters.copy()
            all_bands = list(self.passbands.keys())

            for pb_name in all_bands:
                pb_obj = self.passbands[pb_name]
                if pb_name in filters:
                    filters_remaining.discard(pb_name)
                elif pb_obj.filter_name in filters:
                    filters_remaining.discard(pb_obj.filter_name)
                else:
                    del self.passbands[pb_name]

            if len(filters_remaining) != 0:
                raise ValueError(f"The following filters were not found: {filters_remaining}")

        # Compute the unique points and bounds for the group.
        self._update_internal_data()

    def __str__(self) -> str:
        """Return a string representation of the PassbandGroup."""
        return f"PassbandGroup containing {len(self.passbands)} passbands: {', '.join(self.passbands.keys())}"

    def __len__(self) -> int:
        return len(self.passbands)

    def __getitem__(self, key):
        """Return the passband corresponding to a full name or filter name."""
        if key in self.passbands:
            return self.passbands[key]
        elif key in self._filter_to_name:
            # If we are looking up the passband by filter name, we check
            # that the filter only appears in a single Passband object.
            pb_list = self._filter_to_name[key]
            if len(pb_list) > 1:
                raise KeyError(
                    f"Filter {key} corresponds to multiple passbands: {pb_list}.\n"
                    "Lookup the passband by full name."
                )
            return self.passbands[pb_list[0]]
        else:
            raise KeyError(f"Unknown passband {key}")

    def __contains__(self, key):
        if key in self.passbands:
            return True
        elif key in self._filter_to_name:
            return True
        return False

    @property
    def filters(self) -> list:
        """Return a list of filter names in the passband group."""
        return list(self._filter_to_name.keys())

    @classmethod
    def from_dir(
        cls,
        dir_path: Union[str, Path],
        filters: list | None = None,
        delta_wave: float | None = 5.0,
        trim_quantile: float | None = 1e-3,
        units: Literal["nm", "A"] | None = "A",
    ):
        """Load the passbands from a directory where the directorty name corresponds
        to the survey and the file names correspond to the filters:
        path_to_survey_dir/survey_name/filter_name.dat

        Parameters
        ----------
        dir_path : str or Path
            The path to the passband files including the survey directory.
        filters : list, optional
            A list of filters to include in this PassbandGroup. If None, includes all filters.
            Otherwise drops filters that do not occur and throws an error if a filter is missing.
            Used for loading a subset of the filters.
        delta_wave : float or None, optional
            The grid step of the wave grid, in angstroms.
            It is typically used to downsample transmission using linear interpolation.
            Default is 5 angstroms. If None the original grid is used.
        trim_quantile : float or None, optional
            The quantile to trim the transmission table by. For example, if trim_quantile is 1e-3, the
            transmission table will be trimmed to include only the central 99.8% of the area under the
            transmission curve.
        units : Literal['nm','A'], optional
            Denotes whether the wavelength units of the table are nanometers ('nm') or Angstroms ('A').
            By default 'A'. Does not affect the output units of the class, only the interpretation of the
            provided passband table.
        """
        dir_path = Path(dir_path)
        if not dir_path.is_dir():
            raise ValueError(f"{dir_path} is not a valid directory.")

        # Iterate through the files in the directory.
        all_params = []
        for entry in dir_path.iterdir():
            if entry.is_file():
                filter_name = entry.stem

                # If a list of filters is provided, skip the ones that are not in the list.
                if filters is None or filter_name in filters:
                    params = {
                        "survey": dir_path.name,
                        "filter_name": filter_name,
                        "table_path": dir_path / entry,
                        "delta_wave": delta_wave,
                        "trim_quantile": trim_quantile,
                        "units": units,
                    }
                    all_params.append(params)

        return PassbandGroup(given_passbands=all_params, filters=filters)

    @classmethod
    def from_preset(cls, preset: str, table_dir=None, filters=None, **kwargs) -> None:
        """Create a passband group from a pre-defined set of passbands.

        Parameters
        ----------
        preset : str
            The name of the pre-defined set of passbands to load.
        table_dir : str, optional
            The path to the directory in which to store cached passband tables. If the passband
            exists in this directory, it will be loaded from there; otherwise it will be downloaded
            and saved in that directory.
            The full path to the tables will be {table_dir}/{survey}/{filter_name}.dat.
        filters : list, optional
            A list of filters to include in this PassbandGroup. If None, includes all filters.
            Otherwise drops filters that do not occur and throws an error if a filter is missing.
            Used for loading a subset of the filters.
        **kwargs
            Additional keyword arguments to pass to the Passband constructor.
        """
        logger.info(f"Loading passbands from preset {preset}")
        passbands = []

        if preset == "LSST":
            # Check that units are what is expected for this preset.
            if "units" not in kwargs:
                kwargs["units"] = "nm"
            elif kwargs["units"] != "nm":
                raise ValueError(
                    "LSST passbands are expected to be in nanometers (nm). "
                    "Please set units='nm' in the kwargs."
                )

            if table_dir is None:
                table_dir = Path(_TDASTRO_BASE_DATA_DIR, "passbands", "LSST")
            else:
                table_dir = Path(table_dir) / "LSST"

            for filter_name in ["u", "g", "r", "i", "z", "y"]:
                url = (
                    f"https://github.com/lsst/throughputs/blob/main/baseline/total_{filter_name}.dat?raw=true"
                )
                pb = Passband.from_file(
                    "LSST",
                    filter_name,
                    table_path=Path(table_dir, f"{filter_name}.dat"),
                    table_url=url,
                    **kwargs,
                )
                passbands.append(pb)
        elif preset == "ZTF":
            for filter_name in ["g", "r", "i"]:
                sn_pb = get_bandpass(f"ztf{filter_name}")
                pb = Passband.from_sncosmo("ZTF", filter_name, sn_pb, **kwargs)
                passbands.append(pb)
        else:
            raise ValueError(f"Unknown passband preset: {preset}")

        # Remove any filters that were not in the list to load.
        if filters is not None:
            passbands = [pb for pb in passbands if pb.filter_name in filters]

        # Build the actual PassbandGroup object.
        return cls(given_passbands=passbands, filters=filters, **kwargs)

    def add_passband(self, passband) -> None:
        """Manually add a passband to the group.

        Parameters
        ----------
        passband : Passband
            The passband to add to the group.
        """
        self.passbands[passband.full_name] = passband
        self._update_internal_data()

    def _update_internal_data(self) -> None:
        """Update the cached internal data."""
        # Update the mapping of filter name to full_name.
        self._filter_to_name = {}
        for full_name, pb_obj in self.passbands.items():
            if pb_obj.filter_name in self._filter_to_name:
                logger.info("Multiple passband objects detected for filter {pb_obj.filter_name}")
                self._filter_to_name[pb_obj.filter_name].append(full_name)
            else:
                self._filter_to_name[pb_obj.filter_name] = [full_name]

        self._update_waves()

    def _update_waves(self, threshold=1e-5) -> None:
        """Update the group's wave attribute to be the union of all wavelengths in
        the passbands and update the group's _in_band_wave_indices attribute, which is
        the indices of the group's wave grid that are in the passband's wave grid.

        Eg, if a group's waves are [11, 12, 13, 14, 15] and a single band's are [13, 14],
        we get [2, 3].

        The indices are stored in the passband's _in_band_wave_indices attribute as either
        a tuple of two ints (lower, upper) or a 1D np.ndarray of ints.

        Parameters
        ----------
        threshold : float
            The threshold for merging two "close" wavelengths. This is used to
            avoid problems with numerical precision.
            Default: 1e-5
        """
        if len(self.passbands) == 0:
            self.waves = np.array([])
        else:
            # Compute the unique wavelengths (accounting for floating point error) by
            # sorting the union of all the waves, computing the gaps between each adjacent
            # pair (using 1e8 for before the first), and saving the points with a gap
            # larger than the threshold.
            all_waves = np.concatenate([passband.waves for passband in self.passbands.values()])
            sorted_waves = np.sort(all_waves)
            gap_sizes = np.insert(sorted_waves[1:] - sorted_waves[:-1], 0, 1e8)
            self.waves = sorted_waves[gap_sizes >= threshold]

        # Update the mapping of each passband's wavelengths to the corresponding indices in the
        # unioned list of all wavelengths.
        self._in_band_wave_indices = {}
        for name, passband in self.passbands.items():
            # We only want the fluxes that are in the passband's wavelength range
            # So, find the indices in the group's wave grid that are in the passband's wave grid
            lower, upper = passband.waves[0], passband.waves[-1]
            lower_index, upper_index = np.searchsorted(self.waves, [lower, upper])

            # Check that this is the right grid after all (check will fail if passbands overlap and passbands
            # do not happen to be on the same phase of the grid; eg, even if the step is 10, if the first
            # passband starts at 100 and the second at 105, the passbands won't share the same grid)
            if np.array_equal(self.waves[lower_index : upper_index + 1], passband.waves):
                indices = slice(lower_index, upper_index + 1)
            else:
                indices = np.searchsorted(self.waves, passband.waves)
            self._in_band_wave_indices[name] = indices

    def wave_bounds(self):
        """Get the minimum and maximum wavelength for this group.

        Returns
        -------
        min_wave : float
            The minimum wavelength.
        max_wave : float
            The maximum wavelength.
        """
        min_wave = np.min(self.waves)
        max_wave = np.max(self.waves)
        return min_wave, max_wave

    def mask_by_filter(self, filters):
        """Compute a mask for whether a given observations is of interest for
        for a given analysis. For example this could be used to remove unneeded
        observations from an OpSim or other data set.

        Parameters
        ----------
        filters : list-like
            A length T array of filter names or full names.

        Returns
        -------
        mask : numpy.ndarray
            A length T array of Booleans indicating whether the filter is of interest.
        """
        filters = np.asarray(filters)
        full_name_mask = np.isin(filters, list(self.passbands.keys()))
        filter_name_mask = np.isin(filters, list(self._filter_to_name.keys()))
        return full_name_mask | filter_name_mask

    def process_transmission_tables(self, delta_wave: float | None = 5.0, trim_quantile: float | None = 1e-3):
        """Process the transmission tables for all passbands in the group; recalculate group's wave attribute.

        Parameters
        ----------
        delta_wave : float or None, optional
            The grid step of the wave grid. Default is 5.0.
        trim_quantile : float or None, optional
            The quantile to trim the transmission table by. For example, if trim_quantile is 1e-3, the
            transmission table will be trimmed to include only the central 99.8% of rows.
        """
        for passband in self.passbands.values():
            passband.process_transmission_table(delta_wave, trim_quantile)

        self._update_internal_data()

    def fluxes_to_bandflux(self, flux_density_matrix: np.ndarray, filter: str) -> np.ndarray:
        """Calculate bandfluxes for a single passband in the group.

        Parameters
        ----------
        flux_density_matrix : np.ndarray
            A 2D array of flux densities where of shape T x W where the rows are times and
            columns are wavelengths.
        filter : str
            The name of the filter to evaluate.

        Returns
        -------
        bandflux : np.ndarray
            A length T array of bandfluxes for the given passband.
        """
        if filter not in self.passbands:
            if filter in self._filter_to_name:
                filter = self._filter_to_name[filter][0]
            else:
                raise ValueError(f"Filter {filter} not found in passband group.")
        passband = self.passbands[filter]

        # Evaluate the bandflux using only the wavelengths for this passband.
        wave_indices = self._in_band_wave_indices[filter]
        if wave_indices is None:
            raise ValueError(
                f"Passband {filter} does not have _in_band_wave_indices set. "
                "This should have been calculated in PassbandGroup._update_internal_data."
            )
        in_band_fluxes = flux_density_matrix[:, wave_indices]
        bandflux = passband.fluxes_to_bandflux(in_band_fluxes)

        return bandflux

    def fluxes_to_bandfluxes(self, flux_density_matrix: np.ndarray) -> np.ndarray:
        """Calculate bandfluxes for all passbands in the group.

        Parameters
        ----------
        flux_density_matrix : np.ndarray
            A 2D array of flux densities where of shape T x W where the rows are times and
            columns are wavelengths.

        Returns
        -------
        dict of np.ndarray
            A dictionary of bandfluxes with passband full names as keys and np.ndarrays of
            bandfluxes as values.
        """
        if flux_density_matrix.size == 0 or len(self.waves) != len(flux_density_matrix[0]):
            flux_density_matrix_num_cols = 0 if flux_density_matrix.size == 0 else len(flux_density_matrix[0])
            raise ValueError(
                f"PassbandGroup mismatched grids: Flux density matrix has {flux_density_matrix_num_cols} "
                f"columns, which does not match transmission table's {len(self.waves)} rows. Check that the "
                f"flux density matrix was calculated on the same grid as the transmission tables, which can "
                f"be accessed via the Passband's or PassbandGroup's waves attribute."
            )

        # Compute the bandfluxes for each passband.
        bandfluxes = {}
        for full_name in self.passbands:
            bandfluxes[full_name] = self.fluxes_to_bandflux(flux_density_matrix, full_name)
        return bandfluxes

    def plot(self, ax=None, figure=None):
        """Plot the PassbandGroup on a single plot.

        Parameters
        ----------
        ax : matplotlib.pyplot.Axes or None, optional
            Axes, None by default.
        figure : matplotlib.pyplot.Figure or None
            Figure, None by default.
        """
        if ax is None:
            if figure is None:
                figure = plt.figure()
            ax = figure.add_axes([0, 0, 1, 1])

        # Plot each passband.
        for pb_obj in self.passbands.values():
            pb_obj.plot(ax=ax, plot_loaded=False)
        ax.legend()


class Passband:
    """A passband contains information about its transmission curve and calculates its normalization.

    Attributes
    ----------
    survey : str
        The survey to which the passband belongs: eg, "LSST".
    filter_name : str
        The name of the passband's filter: eg, "u".
    full_name : str
        The full name of the passband. This is the survey and filter concatenated: eg, "LSST_u".
    waves : np.ndarray
        The wavelengths of the transmission table in Angstroms. To be used when evaluating models
        to generate fluxes that will be passed to fluxes_to_bandflux.
    _loaded_table : np.ndarray
        A 2D array of wavelengths and transmissions. This is the table loaded from the file, and is neither
        interpolated nor normalized.
    processed_transmission_table : np.ndarray
        A 2D array where the first col is wavelengths (Angstrom) and the second col is transmission values.
        This table is both interpolated to the _wave_grid and normalized to calculate phi_b(λ).
    """

    def __init__(
        self,
        table_values: np.array,
        survey: str,
        filter_name: str,
        delta_wave: float | None = 5.0,
        trim_quantile: float | None = 1e-3,
        units: Literal["nm", "A"] | None = "A",
    ):
        """Construct a Passband object.

        Parameters
        ----------
        table_values : np.ndarray, optional
            A 2D array of wavelengths (in the given units) and transmissions.
        survey : str
            The survey to which the passband belongs: eg, "LSST".
        filter_name : str
            The filter_name of the passband: eg, "u".
        delta_wave : float or None, optional
            The grid step of the wave grid, in angstroms.
            It is typically used to downsample transmission using linear interpolation.
            Default is 5 angstroms. If None the original grid is used.
        trim_quantile : float or None, optional
            The quantile to trim the transmission table by. For example, if trim_quantile is 1e-3, the
            transmission table will be trimmed to include only the central 99.8% of the area under the
            transmission curve.
        units : Literal['nm','A'], optional
            Denotes whether the wavelength units of the table are nanometers ('nm') or Angstroms ('A').
            By default 'A'. Does not affect the output units of the class, only the interpretation of the
            provided passband table.
        """
        self.survey = survey
        self.filter_name = filter_name
        self.full_name = f"{survey}_{filter_name}"

        # Perform validation of the transmission table.
        if table_values.shape[1] != 2:
            raise ValueError("Passband requires an input table with exactly two columns.")
        diffs = np.diff(table_values[:, 0])
        if np.any(diffs < 0.0):
            raise ValueError("Wavelengths in transmission table must be strictly increasing.")
        if np.any(diffs == 0.0):
            logger.warning("Duplicate wavelengths found in transmission table; averaging values.")
            dup_inds = np.where(diffs == 0.0)
            table_values[dup_inds, 1] = 0.5 * (table_values[dup_inds, 1] + table_values[dup_inds + 1, 1])
            table_values = np.delete(table_values, dup_inds + 1, axis=0)
        self._loaded_table = np.copy(table_values)

        # Ensure the wavelengths are in Angstroms.
        if units == "nm":
            # Multiply the first column (wavelength) by 10.0 to convert to Angstroms
            self._loaded_table[:, 0] *= 10.0
        elif units != "A":
            raise ValueError(f"Unknown Passband units {units}")

        # Preprocess the passband.
        self.process_transmission_table(delta_wave, trim_quantile)

    def __str__(self) -> str:
        """Return a string representation of the Passband."""
        return f"Passband: {self.full_name}"

    def __eq__(self, other) -> bool:
        """Determine if two passbands have equal values for the processed tables."""
        # Check that they are using the same wavelengths.
        if len(self.waves) != len(other.waves):
            return False
        if not np.allclose(self.waves, other.waves):
            return False

        # Check that they have the (approximately) same transmission tables.
        if self.processed_transmission_table.shape != other.processed_transmission_table.shape:
            return False
        return np.allclose(self.processed_transmission_table, other.processed_transmission_table)

    @classmethod
    def from_file(
        cls,
        survey: str,
        filter_name: str,
        delta_wave: float | None = 5.0,
        trim_quantile: float | None = 1e-3,
        table_path: Union[str, Path] | None = None,
        table_url: str | None = None,
        units: Literal["nm", "A"] | None = "A",
        force_download: bool = False,
    ):
        """Construct a Passband object from a file, downloading it if needed.

        Parameters
        ----------
        survey : str
            The survey to which the passband belongs: eg, "LSST".
        filter_name : str
            The filter_name of the passband: eg, "u".
        delta_wave : float or None, optional
            The grid step of the wave grid, in angstroms.
            It is typically used to downsample transmission using linear interpolation.
            Default is 5 angstroms. If None the original grid is used.
        trim_quantile : float or None, optional
            The quantile to trim the transmission table by. For example, if trim_quantile is 1e-3, the
            transmission table will be trimmed to include only the central 99.8% of the area under the
            transmission curve.
        table_path : str, optional
            The path to the transmission table file. If None, the table path will be set to a default path;
            if no file exists at this location, the file will be downloaded from table_url.
        table_url : str, optional
            The URL to download the transmission table file. If None, the table URL will be set to
            a default URL based on the survey and filter_name. Default is None.
        units : Literal['nm','A'], optional
            Denotes whether the wavelength units of the table are nanometers ('nm') or Angstroms ('A').
            By default 'A'. Does not affect the output units of the class, only the interpretation of the
            provided passband table.
        force_download : bool, optional
            If True, the transmission table will be downloaded even if it already exists locally. Default is
            False.
        """
        if table_path is None:
            # If no path is given, use the default.
            table_path = Path(
                _TDASTRO_BASE_DATA_DIR,
                "passbands",
                survey,
                f"{filter_name}.dat",
            )
        else:
            table_path = Path(table_path)

        # Download the table if it does not exist or if force_download is True.
        success = download_data_file_if_needed(table_path, table_url, force_download=force_download)
        if not success:
            raise RuntimeError(f"Failed to download passband table from {table_url}.")

        # Load the table and create the passband.
        loaded_table = Passband.load_transmission_table(table_path)
        return Passband(
            loaded_table,
            survey,
            filter_name,
            delta_wave,
            trim_quantile,
            units=units,
        )

    @classmethod
    @cite_function("https://sncosmo.readthedocs.io/en/stable/api/sncosmo.Bandpass.html")
    def from_sncosmo(cls, survey: str, filter_name: str, bandpass: Bandpass):
        """Create a Passband object from an sncosmo.Bandpass object.

        Parameters
        ----------
        survey : str
            The survey to which the passband belongs: eg, "LSST".
        filter_name : str
            The filter_name of the passband: eg, "u".
        bandpass : sncosmo.Bandpass
            The bandpass object from which to create the Passband object.

        Reference
        ---------
        snocosmo.Bandpass:
        https://sncosmo.readthedocs.io/en/stable/api/sncosmo.Bandpass.html
        """
        table = np.column_stack([bandpass.wave, bandpass.trans])
        return Passband(
            table,
            survey,
            filter_name,
            trim_quantile=None,  # Trimming is done in sncosmo
            units="A",  # All sncosmo bandpasses are in Angstroms
        )

    @staticmethod
    def load_transmission_table(table_path: Union[str, Path], **kwargs) -> np.ndarray:
        """Load a transmission table from a file.

        Table must have 2 columns: wavelengths and transmissions; wavelengths must be
        strictly increasing.

        Parameters
        ----------
        table_path : str or Path
            The path to the transmission table file. If no file exists at this location, the
            file will be downloaded from table_url.
        **kwargs : dict
            Additional keyword arguments to pass to the reader method.

        Returns
        -------
        np.ndarray
            A 2D array of wavelengths and transmissions.
        """
        logger.info(f"Loading passband from file: {table_path}")

        table_path = Path(table_path)
        if not table_path.exists():
            raise FileNotFoundError(f"Transmission table not found at {table_path}")

        # Add default delimiter if not provided
        if (table_path.suffix == ".csv" or table_path.suffix == ".ecsv") and "delimiter" not in kwargs:
            kwargs["delimiter"] = ","

        # Load the table.
        loaded_table = np.loadtxt(table_path, **kwargs)

        # Check that the table has the correct shape
        if loaded_table.size == 0 or loaded_table.shape[1] != 2:
            raise ValueError("Transmission table must have exactly 2 columns.")

        # Check that wavelengths are strictly increasing. If there are duplicates then
        # we average the values.
        diffs = np.diff(loaded_table[:, 0])
        if np.any(diffs < 0.0):
            raise ValueError("Wavelengths in transmission table must be increasing.")
        if np.any(diffs == 0.0):
            logger.warning("Duplicate wavelengths found in transmission table; averaging values.")
            dup_inds = np.where(diffs == 0.0)
            loaded_table[dup_inds, 1] = 0.5 * (loaded_table[dup_inds, 1] + loaded_table[dup_inds + 1, 1])
            loaded_table = np.delete(loaded_table, dup_inds + 1, axis=0)

        return loaded_table

    def process_transmission_table(
        self,
        delta_wave: float | None = 5.0,
        trim_quantile: float | None = 1e-3,
    ):
        """Process the transmission table.

        Parameters
        ----------
        delta_wave : Optional[float] = 5.0
            The grid step in Angstroms of the wave grid. Default is 5.0 Angstroms.
        trim_quantile : Optional[float] = 1e-3
            The quantile to trim the transmission table by. For example, if trim_quantile is 1e-3, the
            transmission table will be trimmed to include only the central 99.8% of rows.
        """
        interpolated_table = self._interpolate_transmission_table(self._loaded_table, delta_wave)
        trimmed_table = self._trim_transmission_by_quantile(interpolated_table, trim_quantile)
        self.processed_transmission_table = self._normalize_transmission_table(trimmed_table)

        self.waves = self.processed_transmission_table[:, 0]

    def _interpolate_transmission_table(self, table: np.ndarray, delta_wave: float | None) -> np.ndarray:
        """Interpolate the transmission table to a new wave grid.

        Parameters
        ----------
        table : np.ndarray
            A 2D array of wavelengths (in Angstroms) and transmissions.
        delta_wave : float or None
            The grid step in Angstroms of the wave grid.

        Returns
        -------
        np.ndarray
            The 2D interpolated array of wavelengths (in Angstroms) and transmissions.
        """
        # Don't interpolate if delta_wave is None or the table is already on the desired grid
        if delta_wave is None:
            return table
        if np.allclose(np.diff(table[:, 0]), delta_wave):
            return table

        # Regrid wavelengths to the new wave grid
        wavelengths = table[:, 0]
        lower_bound, upper_bound = wavelengths[0], wavelengths[-1]
        new_wavelengths = np.linspace(
            lower_bound, upper_bound, int((upper_bound - lower_bound) / delta_wave) + 1
        )

        # Interpolate the transmission table to the new wave grid
        spline = scipy.interpolate.InterpolatedUnivariateSpline(table[:, 0], table[:, 1], ext="raise", k=1)
        interpolated_transmissions = spline(new_wavelengths)
        return np.column_stack((new_wavelengths, interpolated_transmissions))

    def _trim_transmission_by_quantile(self, table: np.ndarray, trim_quantile: float | None) -> np.ndarray:
        """Trim the transmission table so that it only includes the central (100 - 2*trim_quartile)% of rows.

        E.g., if trim_quantile is 1e-3, the transmission table will be trimmed to include only the central
        99.8% of rows.

        Parameters
        ----------
        table : np.ndarray
            A 2D array of wavelengths (in Angstroms) and transmissions.
        trim_quantile : float
            The quantile to trim the transmission table by. For example, if trim_quantile is 1e-3, the
            transmission table will be trimmed to include only the central 99.8% of rows. Must be greater than
            or equal to 0 and less than 0.5.

        Returns
        -------
        np.ndarray
            A 2D array of wavelengths (in Angstroms) and transmissions.
        """
        if trim_quantile is None or trim_quantile == 0.0:
            return table
        if trim_quantile < 0 or trim_quantile >= 0.5:
            raise ValueError(f"Trim quantile must be between 0 and 0.5; got {trim_quantile}.")

        # Separate wavelengths and transmissions
        wavelengths = table[:, 0]
        transmissions = table[:, 1]

        # Calculate the cumulative sum of the transmission values (area under the curve)
        cumulative_area = scipy.integrate.cumulative_trapezoid(transmissions, x=wavelengths)

        # Normalize cumulative area to range from 0 to 1
        cumulative_area /= cumulative_area[-1]

        # Find indices where the cumulative area exceeds the trim quantiles
        lower_bound = max(np.searchsorted(cumulative_area, trim_quantile, side="right") - 1, 0)
        upper_bound = np.searchsorted(cumulative_area, 1 - trim_quantile)

        # Trim the table to the desired range
        trimmed_table = table[lower_bound : upper_bound + 1]

        return trimmed_table

    def _normalize_transmission_table(self, transmission_table: np.ndarray) -> np.ndarray:
        """Calculate the value of phi_b for all wavelengths in a transmission table.

        This is eq. 8 from "On the Choice of LSST Flux Units" (Ivezić et al.):

        φ_b(λ) = S_b(λ)λ⁻¹ / ∫ S_b(λ)λ⁻¹ dλ

        where S_b(λ) is the system response of the passband. Note we use transmission table as our S_b(λ).

        Parameters
        ----------
        transmission_table : np.ndarray
            A 2D array of wavelengths (in Angstroms) and transmissions.

        Returns
        -------
        np.ndarray
            A 2D array of wavelengths (in Angstroms) and normalized transmissions.

        Raises
        ------
        ValueError
            If the transmission table is the wrong size or the calculated denominator is zero.
        """
        if transmission_table.size == 0:
            raise ValueError("Transmission table is empty; cannot normalize.")
        elif transmission_table.shape[0] == 1:
            raise ValueError(f"Cannot normalize transmission table with only one row: {transmission_table}.")
        elif transmission_table.ndim != 2 or transmission_table.shape[1] != 2:
            raise ValueError("Transmission table must be 2D array with exactly 2 columns.")

        wavelengths_angstrom = transmission_table[:, 0]
        transmissions = transmission_table[:, 1]
        # Calculate the numerators and denominator
        numerators = transmissions / wavelengths_angstrom
        denominator = scipy.integrate.trapezoid(numerators, x=wavelengths_angstrom)

        if denominator == 0:
            raise ValueError(
                "Denominator is zero; cannot normalize transmission table. "
                f"Consider checking the transmission table for band {self.full_name}. "
                f"With wave grid set to: {self._wave_grid}, transmission table wavelengths are: "
                f"{transmission_table[:, 0]}."
            )

        # Calculate phi_b for each wavelength
        normalized_transmissions = numerators / denominator
        return np.column_stack((wavelengths_angstrom, normalized_transmissions))

    def fluxes_to_bandflux(
        self,
        flux_density_matrix: np.ndarray,
    ) -> np.ndarray:
        """Calculate the bandflux for a given set of flux densities.

        This is eq. 7 from "On the Choice of LSST Flux Units" (Ivezić et al.):

        F_b = ∫ f(λ)φ_b(λ) dλ

        where f(λ) is the flux density of an object at the top of the atmosphere, and φ_b(λ) is the
        normalized system response for given band b."

        Parameters
        ----------
        flux_density_matrix : np.ndarray
            A 2D or 3D array of flux densities. If the array is 2D it contains a single sample where
            the rows are the T times and columns are M wavelengths in Angstroms. If the array is 3D
            it contains S samples and the values are indexed as (sample_num, time, wavelength).

        Returns
        -------
        bandfluxes : np.ndarray
            A 1D or 2D array. If the flux_density_matrix contains a single sample (2D input) then
            the function returns a 1D length T array where each element is the bandflux
            at the corresponding time. Otherwise the function returns a size S x T array where
            each entry corresponds to the value for a given sample at a given time.
        """
        if flux_density_matrix.size == 0:
            raise ValueError("Empty flux density matrix used.")
        if len(flux_density_matrix.shape) == 2:
            w_axis = 1
            flux_density_matrix_num_cols = flux_density_matrix.shape[1]
        elif len(flux_density_matrix.shape) == 3:
            w_axis = 2
            flux_density_matrix_num_cols = flux_density_matrix.shape[2]
        else:
            raise ValueError("Invalid flux density matrix. Must be 2 or 3-dimensional.")

        # Check the number of wavelengths match.
        if len(self.waves) != flux_density_matrix_num_cols:
            raise ValueError(
                f"Passband mismatched grids: Flux density matrix has {flux_density_matrix_num_cols} "
                f"columns, which does not match the {len(self.waves)} rows in band {self.full_name}'s "
                f"transmission table. Check that the flux density matrix was calculated on the same grid as "
                f"the transmission tables, which can be accessed via the Passband's or PassbandGroup's waves "
                f"attribute."
            )

        # Calculate the bandflux as ∫ f(λ)φ_b(λ) dλ,
        # where f(λ) is the flux density and φ_b(λ) is the normalized system response
        integrand = flux_density_matrix * self.processed_transmission_table[:, 1]
        bandfluxes = scipy.integrate.trapezoid(integrand, x=self.waves, axis=w_axis)
        return bandfluxes

    def plot(self, ax=None, figure=None, color=None, plot_loaded=False):
        """Plot the passband.

        Parameters
        ----------
        ax : matplotlib.pyplot.Axes or None, optional
            Axes, None by default.
        figure : matplotlib.pyplot.Figure or None
            Figure, None by default.
        color : str or None, optional
            The color of the curve.
        plot_loaded : bool
            Also plot the loaded table as a dashed line. Used for debugging.
        """
        if ax is None:
            if figure is None:
                figure = plt.figure()
            ax = figure.add_axes([0, 0, 1, 1])

        # If the color is provided, we use that. Otherwise we try
        # the LSST filter colors (or default to black).
        if color is None:
            color = lsst_filter_plot_colors.get(self.filter_name, "black")

        ax.plot(
            self.processed_transmission_table[:, 0],  # X values are the wavelength
            self.processed_transmission_table[:, 1],  # Y values are the transmission values.
            color=color,
            label=self.full_name,
        )
        if plot_loaded:
            ax.plot(
                self._loaded_table[:, 0],  # X values are the wavelength
                self._loaded_table[:, 1],  # Y values are the transmission values.
                color=color,
                linestyle="--",
            )

        ax.set_xlabel(r"Wavelength, $\AA$")
        ax.set_ylabel("Transmission Value")
        ax.set_ylim(0, None)



import os

import jax.numpy as jnp
import pytest
from tdastro.astro_utils.salt2_color_law import (
    _SALT2CL_B,
    _SALT2CL_V,
    _WAVESCALE,
    SALT2ColorLaw,
)


def test_salt2_color_law():
    """Test that we can create and query a simple SALT2ColorLaw object."""
    test_cl = SALT2ColorLaw(_SALT2CL_B, _SALT2CL_V, [0.5, 0.25, 0.1])

    # The coefficients are shifted (with 1 - sum(others) prepended) and zero padded.
    assert jnp.allclose(test_cl.coeffs, jnp.asarray([0.15, 0.5, 0.25, 0.1, 0.0, 0.0, 0.0]))
    assert test_cl.scaled_wave_min == pytest.approx(0.0)
    assert test_cl.scaled_wave_max == pytest.approx(1.0)
    assert test_cl.value_at_min == pytest.approx(0.0)
    assert test_cl.value_at_max == pytest.approx(1.0)
    assert test_cl.deriv_at_min == pytest.approx(0.15)
    assert test_cl.deriv_at_max == pytest.approx(2.3)

    # Test that we can compute the results for some manually specified points.
    half_step = (_SALT2CL_V - _SALT2CL_B) / 2.0
    test_wavelengths = [
        _SALT2CL_B - half_step,  # Shifted and scaled -> -0.5
        _SALT2CL_B,  # Shifted and scaled -> 0.0
        _SALT2CL_B + half_step,  # Shifted and scaled -> 0.5
        _SALT2CL_V,  # Shifted and scaled -> 1.0
        _SALT2CL_V + half_step,  # Shifted and scaled -> 1.5
    ]
    results = test_cl.apply(test_wavelengths)

    assert results[0] == pytest.approx(0.0 + 0.15 * 0.5)
    assert results[1] == pytest.approx(0.0)
    assert results[2] == pytest.approx(-0.2375)
    assert results[3] == pytest.approx(-1.0)
    assert results[4] == pytest.approx(-1.0 - 2.3 * 0.5)


def test_salt2_color_law_load(test_data_dir):
    """Test loading a Salt2ColorLaw object from a file."""
    filename = os.path.join(test_data_dir, "truncated-salt2-h17/salt2_color_correction.dat")
    test_cl = SALT2ColorLaw.from_file(filename)

    expected_coeffs = [1.83205876, -1.33154627, 0.61225710, -0.12117791, 0.00840832, 0.0, 0.0]
    assert jnp.allclose(test_cl.coeffs, jnp.asarray(expected_coeffs))
    assert test_cl.scaled_wave_min == pytest.approx((2800 - _SALT2CL_B) * _WAVESCALE)
    assert test_cl.scaled_wave_max == pytest.approx((9500 - _SALT2CL_B) * _WAVESCALE)



import numpy as np
from tdastro.graph_state import GraphState
from tdastro.math_nodes.distribution_sampler import DistributionSampler


def test_distribution_sampler():
    """Test that we can generate numbers from a given distribution."""
    # Create a bimodal distribution with peaks at 2.0 and 6.0.
    x_value = np.arange(10)
    pdf_values = np.array([0.0, 0.05, 0.4, 0.05, 0.0, 0.05, 0.4, 0.05, 0.0, 0.0])
    sampler_node = DistributionSampler(
        x_value,
        pdf_values,
        seed=100,
        node_label="sampler",
        outputs=["sample"],
    )

    # Test that we generate samples from the given distribution. We extract the samples
    # from the state to check that they are correctly saved.
    num_samples = 10_000
    state = GraphState(num_samples=num_samples)
    _ = sampler_node.compute(state)
    samples = state["sampler"]["sample"]

    # Check that the samples are from the correct range of x.
    assert np.all(samples >= 0.0)
    assert np.all(samples <= 9.0)

    # Check that most of the samples are near the two peaks (within 1.2) and
    # that there are few samples at other points.
    assert np.sum(np.abs(samples - 2.0) <= 1.2) > 0.4 * num_samples
    assert np.sum(np.abs(samples - 6.0) <= 1.2) > 0.4 * num_samples
    assert np.sum(np.abs(samples - 0.0) <= 0.5) < 0.2 * num_samples
    assert np.sum(np.abs(samples - 9.0) <= 0.5) < 0.05 * num_samples
    assert np.sum(np.abs(samples - 4.0) <= 0.5) < 0.05 * num_samples



from __future__ import annotations  # "type1 | type2" syntax in Python <3.9

import numpy as np
import numpy.typing as npt

from tdastro.astro_utils.mag_flux import mag2flux

_lsstcam_extinction_coeff = {
    "u": -0.458,
    "g": -0.208,
    "r": -0.122,
    "i": -0.074,
    "z": -0.057,
    "y": -0.095,
}
"""The extinction coefficients for the LSST filters.

Values are from
https://community.lsst.org/t/release-of-v3-4-simulations/8548/12
Calculated with syseng_throughputs v1.9
"""

_lsstcam_zeropoint_per_sec_zenith = {
    "u": 26.524,
    "g": 28.508,
    "r": 28.361,
    "i": 28.171,
    "z": 27.782,
    "y": 26.818,
}
"""The zeropoints for the LSST filters at zenith

This is magnitude that produces 1 electron in a 1 second exposure,
see _assign_zero_points() docs for more details.

Values are from
https://community.lsst.org/t/release-of-v3-4-simulations/8548/12
Calculated with syseng_throughputs v1.9
"""


# Suppress "no docstring", because we define it via an attribute.
def magnitude_electron_zeropoint(  # noqa: D103
    *,
    band: npt.ArrayLike,
    airmass: npt.ArrayLike,
    exptime: npt.ArrayLike,
    instr_zp: dict[str, float] | None,
    ext_coeff: dict[str, float] | None,
) -> npt.ArrayLike:
    instr_zp = _lsstcam_zeropoint_per_sec_zenith if instr_zp is None else instr_zp
    ext_coeff = _lsstcam_extinction_coeff if ext_coeff is None else ext_coeff

    instr_zp_getter = np.vectorize(instr_zp.get)
    ext_coeff_getter = np.vectorize(ext_coeff.get)

    return instr_zp_getter(band) + ext_coeff_getter(band) * (airmass - 1) + 2.5 * np.log10(exptime)


magnitude_electron_zeropoint.__doc__ = f"""Photometric zeropoint (magnitude that produces 1 electron) for
    LSST bandpasses (v1.9), using a standard atmosphere scaled
    for different airmasses and scaled for exposure times.

    Parameters
    ----------
    band : ndarray of str
        The filter for which to return the photometric zeropoint.
    airmass : ndarray of float
        The airmass at which to return the photometric zeropoint.
    exptime : ndarray of float
        The exposure time for which to return the photometric zeropoint.
    instr_zp : dict[str, float] or None
        The instrumental zeropoint for each bandpass,
        i.e. AB-magnitude that produces 1 electron in a 1-second exposure.
        Keys are the bandpass names, values are the zeropoints.
        If None, the LSST zeropoints are used:
        {_lsstcam_zeropoint_per_sec_zenith}
    ext_coeff : dict[str, float]
        Atmospheric extinction coefficient for each bandpass.
        Keys are the bandpass names, values are the coefficients.
        If None, the LSST coefficients are used:
        {_lsstcam_extinction_coeff}

    Returns
    -------
    ndarray of float
        AB mags that produces 1 electron.

    Notes
    -----
    Typically, zeropoints are defined as the magnitude of a source
    which would produce 1 count in a 1 second exposure -
    here we use *electron* counts, not ADU counts.

    References
    ----------
    Lynne Jones - https://community.lsst.org/t/release-of-v3-4-simulations/8548/12
    """


# Suppress "no docstring", because we define it via an attribute.
def flux_electron_zeropoint(  # noqa: D103
    *,
    instr_zp_mag: dict[str, float] | None,
    ext_coeff: dict[str, float] | None,
    band: npt.ArrayLike,
    airmass: npt.ArrayLike,
    exptime: npt.ArrayLike,
) -> npt.ArrayLike:
    mag_zp_electron = magnitude_electron_zeropoint(
        instr_zp=instr_zp_mag, ext_coeff=ext_coeff, band=band, airmass=airmass, exptime=exptime
    )
    return mag2flux(mag_zp_electron)


flux_electron_zeropoint.__doc__ = f"""Flux (nJy) producing 1 electron.

    Parameters
    ----------
    band : nparray of str
        The filter for which to return the photometric zeropoint.
    airmass : ndarray of float
        The airmass at which to return the photometric zeropoint.
    exptime : ndarray of float
        The exposure time for which to return the photometric zeropoint.
    instr_zp_mag : dict[str, float]
        The instrumental zeropoint for each bandpass in AB magnitudes,
        i.e. the magnitude that produces 1 electron in a 1-second exposure.
        Keys are the bandpass names, values are the zeropoints.
        If None, the LSST zeropoints are used:
        {_lsstcam_zeropoint_per_sec_zenith}
    ext_coeff : dict[str, float]
        Atmospheric extinction coefficient for each bandpass.
        Keys are the bandpass names, values are the coefficients.
        If None, the LSST coefficients are used:
        {_lsstcam_extinction_coeff}

    Returns
    -------
    ndarray of float
        Flux (nJy) per electron.
    """



"""The base EffectModel class used for all effects."""


class EffectModel:
    """A physical or systematic effect to apply to an observation.

    Effects are not ParameterizedNodes but can have arguments that are
    ParameterizedNodes. All settable parameters of the EffectModel
    must be passed as keyword arguments to the apply() method.

    Attributes
    ----------
    rest_frame : bool
        Whether the effect is applied in the rest frame of the observation (True)
        or in the observed frame (False).
    parameters : dict
        A dictionary of parameters for the effect. Maps the parameter names to
        their setters.
    """

    def __init__(self, rest_frame=True, **kwargs):
        self.rest_frame = rest_frame

        self.parameters = {}
        for key, value in kwargs.items():
            self.add_effect_parameter(key, value)

    def add_effect_parameter(self, name, setter):
        """Add a parameter to the effect.

        Note
        ----
        These parameters are automatically added to the corresponding source
        nodes so they are sampled and recorded with the model's other parameters.

        Parameters
        ----------
        name : str
            The name of the parameter.
        setter : function
            A function that sets the parameter value.
        """
        self.parameters[name] = setter

    def apply(self, flux_density, times=None, wavelengths=None, rng_info=None, **kwargs):
        """Apply the effect to observations (flux_density values).

        Parameters
        ----------
        flux_density : numpy.ndarray
            A length T X N matrix of flux density values (in nJy).
        times : numpy.ndarray, optional
            A length T array of times (in MJD).
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : `dict`, optional
           Any additional keyword arguments. This includes all of the
           parameters needed to apply the effect.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of flux densities after the effect is applied (in nJy).
        """
        raise NotImplementedError()



"""Basic functions for downloading data using pooch."""

import logging
from pathlib import Path

import pooch

logger = logging.getLogger(__name__)


def download_data_file_if_needed(data_path, data_url, force_download=False):
    """Download a data file from a URL and save it to a specified path.

    Parameters
    ----------
    data_path : str or Path
        The path to the data file. This is where the downloaded file will be written.
    data_url : str
        The URL to download the data file.
    force_download : bool, optional
        If True, the file will be downloaded even if it already exists. Default is False.

    Returns
    -------
    bool
        True if the download was successful, False otherwise.
    """
    # Start by checking if the file already exists and if we are not forcing a download.
    data_path = Path(data_path)
    if not force_download and data_path.exists():
        logger.info(f"Data file {data_path} already exists. Skipping download.")
        return True

    # Check that there is a valid URL for the download.
    if data_url is None or len(data_url) == 0:
        raise ValueError("No URL given for table download.")
    logger.info(f"Downloading data file from {data_url} to {data_path}")

    # Create the directory in which to save the file if it does not already exist.
    data_path.parent.mkdir(parents=True, exist_ok=True)

    # Use pooch to download the data files and extract them to the data directory.
    full_path = pooch.retrieve(
        url=data_url,
        known_hash=None,
        fname=data_path.name,
        path=data_path.parent,
    )

    if full_path is None or not Path(full_path).exists():
        logger.error(f"Transmission table not downloaded from {data_url}.")
        return False
    return True



"""A model for an AGN.

Adapted from https://github.com/RickKessler/SNANA/blob/master/src/gensed_AGN.py
with the authors' permission.
"""

from os import urandom

import numpy as np
from astropy import constants
from citation_compass import cite_function
from scipy import integrate

from tdastro.base_models import FunctionNode
from tdastro.consts import M_SUN_G
from tdastro.sources.physical_model import PhysicalModel


def sample_damped_random_walk(times, tau_v, sf_inf, t0, rng=None):
    """Sample a damped random walk.

    Parameters
    ----------
    times : np.ndarray
        A length T array with the times at which to sample the damped random
        walk (in MJD).
    tau_v : np.ndarray
        A length W array with timescale of the damped random walk (in s).
    sf_inf : np.ndarray
        A length W array with structure function at infinity time in magnitude.
    t0 : float
        The initial time moment (in MJD).
    rng : np.random.Generator, optional
        The random number generator to use.
        Default: None

    Returns
    -------
    samples : float
        The sampled value of the damped random walk.
    """
    if rng is None:
        rng = np.random.default_rng()

    if len(sf_inf) != len(tau_v):
        raise ValueError(
            "The arrays sf_inf and tau_v must have the same length. "
            f"Received lengths: sf_inf={len(sf_inf)}, tau_v={len(tau_v)}"
        )
    samples = np.zeros((len(times), len(tau_v)))

    # Set the initial values.
    curr_t = t0
    delta_m = rng.random() * sf_inf

    # Iterate over each time step.
    for idx, time in enumerate(times):
        if time <= curr_t and idx != 0:
            raise ValueError("Times must be monotonically increasing.")

        dt = time - curr_t
        delta_m = delta_m * np.exp(-dt / tau_v) + sf_inf * np.sqrt(1 - np.exp(-2 * dt / tau_v)) * rng.random()
        samples[idx, :] = delta_m

    return samples


class AGN(PhysicalModel):
    """A model for an AGN.

    Parameterized values include:
      * accretion_rate - The accretion rate (ME_dot) at Eddington luminosity in g/s.
      * blackhole_accretion_rate - The accretion rate of the black hole in g/s.
      * blackhole_mass - The black hole mass in g.
      * edd_ratio - The Eddington ratio.
      * dec - The object's declination in degrees. [from PhysicalModel]
      * distance - The object's luminosity distance in pc. [from PhysicalModel]
      * L_bol - The bolometric luminosity in erg/s.
      * mag_i - The i band absolute magnitude.
      * ra - The object's right ascension in degrees. [from PhysicalModel]
      * redshift - The object's redshift. [from PhysicalModel]
      * t0 - The t0 of the zero phase, date. [from PhysicalModel]

    Parameters
    ----------
    t0 : float
        initial time moment in days.
    blackhole_mass : float
        The black hole mass in g.
    edd_ratio: float
        Eddington ratio
    seed : int, optional
        The seed to use for the random number generator.
        Default: None
    **kwargs : dict
        Additional keyword arguments.
    """

    def __init__(self, t0, blackhole_mass, edd_ratio, seed=None, **kwargs):
        super().__init__(t0=t0, **kwargs)

        # Add the parameters for the AGN. t0 already set in PhysicalModel.
        self.add_parameter("blackhole_mass", blackhole_mass, **kwargs)
        self.add_parameter("edd_ratio", edd_ratio, **kwargs)

        # Add the derived parameters using FunctionNodes built from the object's static methods.
        # Each of these will be computed for each sample value of the input parameters.
        self.add_parameter(
            "critical_accretion_rate",
            FunctionNode(self.compute_critical_accretion_rate, blackhole_mass=self.blackhole_mass),
            **kwargs,
        )
        self.add_parameter(
            "blackhole_accretion_rate",
            FunctionNode(
                self.compute_blackhole_accretion_rate,
                accretion_rate=self.critical_accretion_rate,  # Pull from computed accretion rate
                edd_ratio=self.edd_ratio,  # Pull from sampled ratio
            ),
            **kwargs,
        )
        self.add_parameter(
            "bolometric_luminosity",
            FunctionNode(
                self.compute_bolometric_luminosity,
                edd_ratio=self.edd_ratio,  # Pull from sampled ratio
                blackhole_mass=self.blackhole_mass,  # Pull from the sampled mass
            ),
            **kwargs,
        )
        self.add_parameter(
            "mag_i",
            FunctionNode(
                self.compute_mag_i,
                bolometric_luminosity=self.bolometric_luminosity,  # Pull from computed value.
            ),
            **kwargs,
        )

        # Set the random number generator for this object.
        if seed is None:
            seed = int.from_bytes(urandom(4), "big")
        self._rng = np.random.default_rng(seed=seed)

    # ------------------------------------------------------------------------
    # --- Static helper methods for computing the derived parameters. --------
    # ------------------------------------------------------------------------

    @staticmethod
    def compute_critical_accretion_rate(blackhole_mass):
        """Compute the critical accretion rate at Eddington luminosity.

        Parameters
        ----------
        blackhole_mass : float
            The black hole mass in g.

        Returns
        -------
        accretion_rate : float
            The accretion rate (ME_dot) at Eddington luminosity in g/s.
        """
        return 1.4e18 * blackhole_mass / M_SUN_G

    @staticmethod
    def compute_blackhole_accretion_rate(accretion_rate, edd_ratio):
        """Compute the accretion rate of the blackhole.

        Parameters
        ----------
        accretion_rate : float
            The accretion rate (ME_dot) at Eddington luminosity in g/s.
        edd_ratio : float
            The Eddington ratio.

        Returns
        -------
        bh_accretion_rate : float
            The accretion rate of the black hole in g/s.
        """
        return accretion_rate * edd_ratio

    @staticmethod
    def compute_bolometric_luminosity(edd_ratio, blackhole_mass):
        """Compute the bolometric luminosity of an AGN.

        Parameters
        ----------
        edd_ratio : float
            The Eddington ratio.
        blackhole_mass : float
            The black hole mass in g.

        Returns
        -------
        bolometric_luminosity : float
            The bolometric luminosity in erg/s.
        """
        return edd_ratio * 1.26e38 * blackhole_mass / M_SUN_G

    @staticmethod
    @cite_function
    def compute_flux_standard_disk(Mdot, nu, rin, i, d, M):
        """Compute the flux based on a standard disk model.

        References
        ----------
        Lipunova, G., Malanchev, K., Shakura, N. (2018)
        https://doi.org/10.1007/978-3-319-93009-1_1

        Parameters
        ----------
        Mdot : float or np.ndarray
            The accretion rate at the previous time step.
        nu : float or np.ndarray
            The frequency.
        rin : float or np.ndarray
            The inner radius of the accretion disk.
        i : float or np.ndarray
            The inclination.
        d : float or np.ndarray
            The distance.
        M : float or np.ndarray
            The mass of the gravitating center.

        Returns
        -------
        flux : float
            The flux at the given time step.
        """
        # Compute the initial radius of the ring (r_0) and the effective temperature at r_0 (T_0).
        # Lipunova, G., Malanchev, K., Shakura, N. (2018). Page 33 for the main equation
        # DOI https://doi.org/10.1007/978-3-319-93009-1_1
        r_0 = AGN.compute_r_0(rin)
        T_0 = AGN.compute_temp_at_r_0(M, Mdot, rin)

        # large x in exponetial causes overflow, but 1/inf is zero.
        with np.errstate(over="ignore"):

            def _fun_integr(x):
                return (x ** (5 / 3)) / np.expm1(x)

            integ, _ = integrate.quad(_fun_integr, 1e-6, np.inf)

        h = constants.h.cgs.value
        k_B = constants.k_B.cgs.value
        c = constants.c.cgs.value
        return (
            (16 * np.pi)
            / (3 * d**2)
            * np.cos(i)
            * (k_B * T_0 / h) ** (8 / 3)
            * h
            * (nu ** (1 / 3))
            / (c**2)
            * (r_0**2)
            * integ
        )

    @staticmethod
    @cite_function
    def compute_mag_i(bolometric_luminosity):
        """Compute the i band magnitude from the bolometric luminosity.

        References
        ----------
        Shen et al., 2013 - https://adsabs.harvard.edu/full/2013BASI...41...61S

        Parameters
        ----------
        bolometric_luminosity : float
            The bolometric luminosity in erg/s.

        Returns
        -------
        mag_i : float
            The i band magnitude.
        """
        # Adpated from Shen et al., 2013: https://adsabs.harvard.edu/full/2013BASI...41...61S
        return 90 - 2.5 * np.log10(bolometric_luminosity)

    @staticmethod
    @cite_function
    def compute_r_0(r_in):
        """Compute the initial radius of the ring (r_0) in a standard disk model
        given the inner radius.

        References
        ----------
        Lipunova, G., Malanchev, K., Shakura, N. (2018)
        https://doi.org/10.1007/978-3-319-93009-1_1

        Parameters
        ----------
        r_in : float
            The inner radius of the accretion disk.

        Returns
        -------
        r_0 : float
            The initial radius of the ring.
        """
        # Adapted from Lipunova, G., Malanchev, K., Shakura, N. (2018). Page 33 for the main equation
        # DOI https://doi.org/10.1007/978-3-319-93009-1_1
        return (7 / 6) ** 2 * r_in

    @staticmethod
    @cite_function
    def compute_structure_function_at_inf(wavelength, mag_i=-23, blackhole_mass=1e9 * M_SUN_G):
        """Compute the structure function at infinity time in magnitude.

        References
        ----------
        Suberlak et al. 2021 - DOI 10.3847/1538-4357/abc698

        Parameters
        ----------
        wavelength : np.ndarray
            A length W array with the frequencies in Hz.
        mag_i : float, optional
            The i band magnitude.
            Default: -23
        blackhole_mass : float, optional
            The black hole mass in g.
            Default: 1e9 * M_SUN_G

        Returns
        -------
        result : np.ndarray
            A length W array with structure function at infinity time in magnitude.
        """
        # Equation and parameters for A=-0.51, B=-0.479, C=0.13, and D=0.18
        #  adopted from Suberlak et al. 2021: DOI 10.3847/1538-4357/abc698
        return 10 ** (
            -0.51
            - 0.479 * np.log10(wavelength / (4000e-8))
            + 0.13 * (mag_i + 23)
            + 0.18 * np.log10(blackhole_mass / (1e9 * M_SUN_G))
        )

    @staticmethod
    @cite_function
    def compute_tau_v_drw(wavelength, mag_i=-23, blackhole_mass=1e9 * M_SUN_G):
        """Compute the timescale (tau_v) for the DRW model.

        References
        ----------
        Suberlak et al. 2021 - DOI 10.3847/1538-4357/abc698

        Parameters
        ----------
        wavelength : np.ndarray
            A length W array with the frequenc in Hz.
        mag_i : float, optional
            The i band magnitude.
            Default: -23
        blackhole_mass : float, optional
            The black hole mass in g.
            Default: 1e9 * M_SUN_G

        Returns
        -------
        tau_v : np.ndarray
            A length W array with the timescale in s for each wavelength.
        """
        # Equation and parameters for A=2.4, B=0.17, C=0.03, and D=0.21 adopted
        # from Suberlak et al. 2021: DOI 10.3847/1538-4357/abc698
        return 10 ** (
            2.4
            + 0.17 * np.log10(wavelength / (4000e-8))
            + 0.03 * (mag_i + 23)
            + 0.21 * np.log10(blackhole_mass / (1e9 * M_SUN_G))
        )

    @staticmethod
    @cite_function
    def compute_temp_at_r_0(M, Mdot, r_in):
        """Compute the effective temperature at r0. This is the same as the maximum effective
        temperature at the disc surface (Tmax).

        References
        ----------
        Lipunova, G., Malanchev, K., Shakura, N. (2018)
        https://doi.org/10.1007/978-3-319-93009-1_1

        Parameters
        ----------
        M : float
            The mass of the gravitating centre.
        Mdot : float
            The accretion rate at the previous time step.
        r_in : float
            The inner radius of the accretion disk.

        Returns
        -------
        T_0 : float
            The effective temperature at r0.
        """
        # Lipunova, G., Malanchev, K., Shakura, N. (2018). Page 33 for the main equation
        # DOI https://doi.org/10.1007/978-3-319-93009-1_1
        sigma_sb = constants.sigma_sb.cgs.value
        G = constants.G.cgs.value
        return 2 ** (3 / 4) * (3 / 7) ** (7 / 4) * (G * M * Mdot / (np.pi * sigma_sb * r_in**3)) ** (1 / 4)

    def compute_flux(self, times, wavelengths, graph_state, **kwargs):
        """Draw effect-free observations for this object.

        Parameters
        ----------
        times : `numpy.ndarray`
            A length T array of rest frame timestamps.
        wavelengths : `numpy.ndarray`, optional
            A length N array of wavelengths (in angstroms).
        graph_state : `GraphState`
            An object mapping graph parameters to their values.
        **kwargs : `dict`, optional
            Any additional keyword arguments.

        Returns
        -------
        flux_density : `numpy.ndarray`
            A length T x N matrix of SED values (in nJy).
        """
        params = self.get_local_params(graph_state)

        # Compute the parameters for these wavelengths.
        tau_v = self.compute_tau_v_drw(wavelengths, params["mag_i"], params["blackhole_mass"])
        sf_inf = self.compute_structure_function_at_inf(
            wavelengths, params["mag_i"], params["blackhole_mass"]
        )

        # Compute the average flux of a standard disk model. Use a factor of 2 (two sides
        # of the disk) to get the total flux.
        fnu_average = 2.0 * self.compute_flux_standard_disk(
            params["blackhole_accretion_rate"],
            constants.c.cgs.value / wavelengths,  # nu
            1,  # rin
            0,  # i
            1.0,  # d
            params["blackhole_mass"],
        )

        # Run the damped random walk.
        delta_m = sample_damped_random_walk(
            times,
            tau_v,
            sf_inf,
            params["t0"],
            rng=self._rng,
        )

        flux_density = 10 ** (-0.4 * delta_m) * fnu_average
        return flux_density



"""Benchmarks for core TDAstro functionality.

To manually run the benchmarks use: asv run

For more information on writing benchmarks:
https://asv.readthedocs.io/en/stable/writing_benchmarks.html."""

from pathlib import Path

import numpy as np
from astropy import units as u
from tdastro.astro_utils.passbands import PassbandGroup
from tdastro.astro_utils.snia_utils import DistModFromRedshift, HostmassX1Func, X0FromDistMod
from tdastro.astro_utils.unit_utils import fnu_to_flam
from tdastro.base_models import FunctionNode
from tdastro.effects.white_noise import WhiteNoise
from tdastro.math_nodes.np_random import NumpyRandomFunc
from tdastro.sources.basic_sources import LinearWavelengthSource, StepSource
from tdastro.sources.lightcurve_source import LightcurveSource
from tdastro.sources.sncomso_models import SncosmoWrapperModel

# ASV runs from copy of the project (benchmarks/env/....). So we load the
# data files based off the current file location instead.
_PROJECT_BASE_DIR = Path(__file__).parent.parent
_TEST_DATA_DIR = _PROJECT_BASE_DIR / "tests" / "tdastro" / "data"


def _load_test_passbands():
    """Load passbands to use in various benchmarks."""
    passbands_dir = _TEST_DATA_DIR / "passbands"
    passbands = PassbandGroup(
        given_passbands=[
            {
                "filter_name": "g",
                "survey": "LSST",
                "table_path": passbands_dir / "LSST" / "g.dat",
            },
            {
                "filter_name": "r",
                "survey": "LSST",
                "table_path": passbands_dir / "LSST" / "r.dat",
            },
        ],
        survey="LSST",
        units="nm",
        trim_quantile=0.001,
        delta_wave=1,
    )
    return passbands


class TimeSuite:
    """A suite of timing functions."""

    def setup(self):
        """Set up items that will be used in multiple tests."""
        # Preload the passbands for tests that use them.
        self.passbands = _load_test_passbands()

        # Create a model we can use in tests.
        self.redshift = 0.1
        self.hostmass = 8.0
        self.distmod_func = DistModFromRedshift(self.redshift, H0=73.0, Omega_m=0.3)
        self.x1_func = HostmassX1Func(self.hostmass)
        self.x0_func = X0FromDistMod(
            distmod=self.distmod_func,
            x1=self.x1_func,
            c=0.0,
            alpha=0.14,
            beta=3.1,
            m_abs=-19.3,
        )

        self.salt3_model = SncosmoWrapperModel(
            "salt3",
            t0=0.0,
            x0=self.x0_func,
            x1=self.x1_func,
            c=0.0,
            ra=0.0,
            dec=0.0,
            redshift=self.redshift,
        )

        # A simple LinearWavelengthSource that we can use in tests.
        self.linear_source = LinearWavelengthSource(linear_base=1.0, linear_scale=0.1)

        # Create samples that we can use in tests.
        self.times = np.arange(-20.0, 50.0, 0.5)
        self.wavelengths = self.passbands.waves
        self.filter_options = ["LSST_g", "LSST_r"]
        self.filters = np.array([self.filter_options[i % 2] for i in range(len(self.times))])

        self.graph_state = self.salt3_model.sample_parameters()
        self.fluxes = self.salt3_model.evaluate(self.times, self.wavelengths, graph_state=self.graph_state)

        self.white_noise = WhiteNoise(white_noise_sigma=0.1)

    def time_chained_evaluate(self):
        """Time the generation of random numbers with an numpy generation node."""

        def _add_func(a, b):
            return a + b

        # Generate a starting mean and scale from uniform distributions. Use those to
        # generate a sample from the normal distribution. Then shift that sample by -5.0.
        loc_node = NumpyRandomFunc("uniform", low=10.0, high=20.0)
        scale_node = NumpyRandomFunc("uniform", low=0.5, high=1.0)
        norm_node = NumpyRandomFunc("normal", loc=loc_node, scale=scale_node)
        val_node = FunctionNode(_add_func, a=norm_node, b=-5.0)

        # Generate 100,000 samples.
        _ = val_node.sample_parameters(num_samples=100_000)

    def time_apply_white_noise(self):
        """Time the application of white noise to a sample."""
        _ = self.white_noise.apply(self.fluxes, white_noise_sigma=0.1)

    def time_make_x1_from_hostmass(self):
        """Time the creation of the X1 function."""
        _ = HostmassX1Func(self.hostmass)

    def time_sample_x1_from_hostmass(self):
        """Time the computation of the X1 function."""
        _ = self.x1_func.sample_parameters()

    def time_sample_x0_from_distmod(self):
        """Time the computation of the X0 function."""
        _ = self.x0_func.sample_parameters()

    def time_make_and_evaluate_step_source(self):
        """Time creating and evaluating a StepSource."""
        model = StepSource(brightness=100.0, t0=2.0, t1=5.0)
        state = model.sample_parameters()
        times = np.arange(0.0, 10.0, 0.05)
        wavelengths = np.arange(1000.0, 2000.0, 5.0)
        _ = model.evaluate(times, wavelengths, state)

    def time_make_simple_linear_wavelength_source(self):
        """Time creating a simple LinearWavelengthSource."""
        _ = LinearWavelengthSource(linear_base=1.0, linear_scale=0.1)

    def time_evaluate_simple_linear_wavelength_source(self):
        """Time evaluating a simple LinearWavelengthSource."""
        _ = self.linear_source.evaluate(self.times, self.wavelengths)

    def time_make_new_salt3_model(self):
        """Time creating a new SALT3 model."""
        _ = SncosmoWrapperModel(
            "salt3",
            t0=0.0,
            x0=self.x0_func,
            x1=self.x1_func,
            c=0.0,
            ra=0.0,
            dec=0.0,
            redshift=self.redshift,
        )

    def time_evaluate_salt3_model(self):
        """Time querying a predefined salt3 model."""
        _ = self.salt3_model.evaluate(
            self.times,
            self.wavelengths,
            graph_state=self.graph_state,
        )

    def time_load_passbands(self):
        """Time loading the passbands from files."""
        _ = _load_test_passbands()

    def time_apply_passbands(self):
        """Time applying the (already loaded) passbands to flux."""
        _ = self.passbands.fluxes_to_bandfluxes(self.fluxes)

    def time_evaluate_salt3_passbands(self):
        """Time evaluate the SALT3 model at the passband level."""
        _ = self.salt3_model.get_band_fluxes(
            self.passbands,
            self.times,
            self.filters,
            self.graph_state,
        )

    def time_fnu_to_flam(self):
        """Time the fnu_to_flam function."""
        _ = fnu_to_flam(
            self.fluxes,
            self.wavelengths,
            wave_unit=u.AA,
            flam_unit=u.erg / u.second / u.cm**2 / u.AA,
            fnu_unit=u.nJy,
        )

    def time_lightcurve_source(self):
        """Time the creation and query of a LightcurveSource."""
        lc_times = np.linspace(0.0, 6 * np.pi, 100)
        g_gluxes = np.sin(lc_times) + 3.0
        r_gluxes = np.cos(lc_times) + 5.0

        lightcurves = {
            "g": np.column_stack((lc_times, g_gluxes)),
            "r": np.column_stack((lc_times, r_gluxes)),
        }
        lc_source = LightcurveSource(
            lightcurves,
            self.passbands,
            lc_t0=0.0,
            periodic=True,
            baseline=None,
            t0=0.0,
        )

        # Sample the lightcurve source to ensure it works.
        _ = lc_source.evaluate(self.times, self.wavelengths)



import numpy as np
import pytest
from tdastro.effects.effect_model import EffectModel


def test_effect_model() -> None:
    """The base effect model is able to extract the parameters."""
    model = EffectModel(param1=1.0, param2=2.0)
    assert "param1" in model.parameters
    assert "param2" in model.parameters
    assert model.parameters["param1"] == 1.0
    assert model.parameters["param2"] == 2.0

    # We cannot call apply on the base EffectModel
    with pytest.raises(NotImplementedError):
        _ = model.apply(np.zeros((5, 3)))



import numpy as np
from tdastro.math_nodes.np_random import NumpyRandomFunc
from tdastro.math_nodes.scipy_random import (
    NumericalInversePolynomialFunc,
    SampleLogPDF,
    SamplePDF,
)


class FlatDist:
    """A flat (uniform) distribution used for testing.

    Attributes
    ----------
    min_val : `float`
        The minimum value of the distribution.
    max_val : `float`
        The maximum value of the distribution.
    width : `float`
        The width of the distribution.
    """

    def __init__(self, min_val, max_val):
        if min_val >= max_val:
            raise ValueError("Invalid flat_dist bounds.")
        self.min_val = min_val
        self.max_val = max_val
        self.width = self.max_val - self.min_val

    def pdf(self, xval):
        """The pdf of the distribution.

        Parameters
        ----------
        xval: `float`
            The x value to evaluate.

        Returns
        -------
        result : `float`
            The value of the pdf function at the given points.
        """
        if xval <= self.min_val:
            return 0.0
        elif xval >= self.max_val:
            return 0.0
        else:
            return 1.0 / self.width


def test_numerical_inverse_polynomial_func_object():
    """Test that we can generate numbers from a uniform distribution."""
    dist = FlatDist(min_val=0.25, max_val=0.75)
    scipy_node = NumericalInversePolynomialFunc(dist, seed=100)

    # Test that we get a uniform distribution.
    num_bins = 10
    num_samples = 10_000
    counts = [0] * num_bins
    all_values = [0] * num_samples
    for i in range(10_000):
        value = scipy_node.generate()
        assert 0.25 <= value <= 0.75

        all_values[i] = value
        bin = int((value - 0.25) * (num_bins / 0.5))
        counts[bin] += 1

    for idx in range(num_bins):
        assert np.abs(counts[idx] - 10_000 / num_bins) < 200.0

    # If we reuse the same seed, we get the same values.
    dist2 = FlatDist(min_val=0.25, max_val=0.75)
    scipy_node2 = NumericalInversePolynomialFunc(dist2, seed=100)
    all_values2 = [0] * num_samples
    for i in range(10_000):
        all_values2[i] = scipy_node2.generate()
    assert np.allclose(all_values, all_values2)

    # If we use a different seed, we get different values.
    dist3 = FlatDist(min_val=0.25, max_val=0.75)
    scipy_node3 = NumericalInversePolynomialFunc(dist3, seed=101)
    all_values3 = [0] * num_samples
    for i in range(10_000):
        all_values3[i] = scipy_node3.generate()
    assert not np.allclose(all_values, all_values3)

    # We can batch generate results.
    state = scipy_node.sample_parameters(num_samples=500)
    samples = scipy_node.get_param(state, "function_node_result")
    assert len(samples) == 500
    assert len(np.unique(samples)) > 50
    assert np.abs(np.mean(samples) - 0.5) < 0.01


def test_numerical_inverse_polynomial_func_object_seed():
    """Test that we can generate numbers from a uniform distribution with a given seed."""
    dist = FlatDist(min_val=0.25, max_val=0.75)
    scipy_node = NumericalInversePolynomialFunc(dist)

    # Sample without a given seed.
    state1 = scipy_node.sample_parameters(num_samples=10)
    values1 = scipy_node.get_param(state1, "function_node_result")

    # Re-sample with a given seed=100.
    state2 = scipy_node.sample_parameters(num_samples=10, rng_info=np.random.default_rng(seed=100))
    values2 = scipy_node.get_param(state2, "function_node_result")

    # Re-sample again with a given seed=100.
    state3 = scipy_node.sample_parameters(num_samples=10, rng_info=np.random.default_rng(seed=100))
    values3 = scipy_node.get_param(state3, "function_node_result")

    assert np.allclose(values2, values3)
    assert not np.allclose(values1, values2)
    assert not np.allclose(values1, values3)


def test_numerical_inverse_polynomial_func_class():
    """Test that we can generate numbers from a uniform distribution."""
    scipy_node = NumericalInversePolynomialFunc(
        FlatDist,
        min_val=NumpyRandomFunc("uniform", low=0.0, high=0.4, seed=101),
        max_val=NumpyRandomFunc("uniform", low=0.6, high=1.0, seed=102),
        seed=100,
    )

    # Check that when we sample both the values and the distribution's parameters.
    num_samples = 1000
    states = scipy_node.sample_parameters(num_samples=num_samples)
    values = scipy_node.get_param(states, "function_node_result")
    min_vals = scipy_node.get_param(states, "min_val")
    max_vals = scipy_node.get_param(states, "max_val")
    assert len(np.unique(values)) == num_samples
    assert len(np.unique(min_vals)) == num_samples
    assert len(np.unique(max_vals)) == num_samples

    # Test that the generated values are consistent with the distributions.
    assert np.all(values >= min_vals)
    assert np.all(values <= max_vals)


def test_numerical_sample_pdf():
    """Test that we can create a SamplePDF node from a function."""

    def _triangle_pdf(x):
        if x < 0.0:
            return 0.0
        if x > 2.0:
            return 0.0
        return 1.0 - 0.5 * x

    scipy_node = SamplePDF(_triangle_pdf, seed=100)

    # Test that we get distribution that ramps down as x increases.
    num_samples = 50_000
    counts = np.zeros(10)
    for _ in range(num_samples):
        value = scipy_node.generate()
        assert 0.0 <= value <= 2.0

        bin = int(10 * value / 2.0)
        counts[bin] += 1

    mid_heights = np.array([0.95, 0.85, 0.75, 0.65, 0.55, 0.45, 0.35, 0.25, 0.15, 0.05])
    expected = mid_heights * 0.2 * num_samples
    for idx in range(10):
        assert np.abs(counts[idx] - expected[idx]) < 200.0


def test_numerical_sample_logpdf():
    """Test that we can create a SampleLogPDF node from a function."""

    def _triangle_logpdf(x):
        if x < 0.0:
            return -np.inf
        if x > 2.0:
            return -np.inf
        return np.log(1.0 - 0.5 * x)

    scipy_node = SampleLogPDF(_triangle_logpdf, seed=100)

    # Test that we get distribution that ramps down as x increases.
    num_samples = 50_000
    counts = np.zeros(10)
    for _ in range(num_samples):
        value = scipy_node.generate()
        assert 0.0 <= value <= 2.0

        bin = int(10 * value / 2.0)
        counts[bin] += 1

    mid_heights = np.array([0.95, 0.85, 0.75, 0.65, 0.55, 0.45, 0.35, 0.25, 0.15, 0.05])
    expected = mid_heights * 0.2 * num_samples
    for idx in range(10):
        assert np.abs(counts[idx] - expected[idx]) < 200.0



"""Multiple source models wrap multiple PhysicalModels, allowing the
user to define such operations as additive models, where each source
contributes to the total flux density, or random source models, where
only one source is selected at random for each flux calculation.
"""

import numpy as np

from tdastro.graph_state import GraphState
from tdastro.math_nodes.given_sampler import GivenValueSampler
from tdastro.sources.physical_model import PhysicalModel


class MultiSourceModel(PhysicalModel):
    """A MultiSourceModel wraps multiple PhysicalModels.

    All rest frame effects are applied to each source, allowing different redshifts
    for each source (for unresolved sources).  The observer frame effects are applied
    to the weighted sum of the sources.

    Note: Each source may have its own sampled (RA, dec) position, which are not
    required to align.

    Attributes
    ----------
    sources : list
        A list of PhysicalModel objects to use in the flux calculation.
    num_sources : int
        The number of sources in the model.

    Parameters
    ----------
    sources : list
        A list of PhysicalModel objects to use in the flux calculation.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(
        self,
        sources,
        **kwargs,
    ):
        super().__init__(**kwargs)

        # Check that all sources are PhysicalModel objects and they do not contain
        # observer frame effects.
        for source in sources:
            if not isinstance(source, PhysicalModel):
                raise ValueError("All sources must be PhysicalModel objects.")
            if len(source.obs_frame_effects) > 0:
                raise ValueError("A MultiSourceModel cannot contain sources with observer frame effects.")
        self.sources = sources
        self.num_sources = len(sources)

    def set_graph_positions(self, seen_nodes=None):
        """Force an update of the graph structure (numbering of each node).

        Parameters
        ----------
        seen_nodes : set, optional
            A set of nodes that have already been processed to prevent infinite loops.
            Caller should not set.
        """
        if seen_nodes is None:
            seen_nodes = set()

        # Set the graph positions for this node and each source.
        super().set_graph_positions(seen_nodes=seen_nodes)
        for source in self.sources:
            source.set_graph_positions(seen_nodes=seen_nodes)

    def set_apply_redshift(self, apply_redshift):
        """Toggles the apply_redshift setting.

        Parameters
        ----------
        apply_redshift : bool
            The new value for apply_redshift.
        """
        for source in self.sources:
            source.set_apply_redshift(apply_redshift)

    def add_effect(self, effect):
        """Add an effect to the model.  Rest frame effects are applied to each source
        and observer frame effects are stored in this model.

        Parameters
        ----------
        effect : Effect
            The effect to add to the model.
        """
        if effect.rest_frame:
            # The rest frame effects are applied to each source.
            for source in self.sources:
                source.add_effect(effect)
        else:
            # Add observer frame effects to this model.
            # Add any effect parameters that are not already in the model.
            for param_name, setter in effect.parameters.items():
                if param_name not in self.setters:
                    self.add_parameter(param_name, setter, allow_gradient=False)
            self.obs_frame_effects.append(effect)

    def sample_parameters(self, given_args=None, num_samples=1, rng_info=None):
        """Sample the model's underlying parameters if they are provided by a function
        or ParameterizedModel.

        Parameters
        ----------
        given_args : dict, optional
            A dictionary representing the given arguments for this sample run.
            This can be used as the JAX PyTree for differentiation.
        num_samples : int
            A count of the number of samples to compute.
            Default: 1
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            All the keyword arguments, including the values needed to sample
            parameters.

        Returns
        -------
        graph_state : GraphState
            An object mapping graph parameters to their values.
        """
        # If the graph has not been sampled ever, update the node positions for
        # every node (model, background, effects).
        if self.node_pos is None:
            self.set_graph_positions()

        # We use the same seen_nodes for all sampling calls so each node
        # is sampled at most one time regardless of link structure.
        graph_state = GraphState(num_samples)
        if given_args is not None:
            graph_state.update(given_args, all_fixed=True)

        seen_nodes = {}
        for source in self.sources:
            source._sample_helper(graph_state, seen_nodes, rng_info=rng_info)
        self._sample_helper(graph_state, seen_nodes, rng_info=rng_info)

        return graph_state

    def _evaluate_single(self, times, wavelengths, state, rng_info=None, **kwargs):
        """Evaluate the model and apply the effects for a single, given graph state.
        This function applies redshift, computes the flux density for the object,
        applies rest frames effects, performs the redshift correction (if needed),
        and applies the observer frame effects.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        wavelengths : numpy.ndarray
            A length N array of wavelengths (in angstroms).
        state : GraphState
            An object mapping graph parameters to their values with num_samples=1.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            All the other keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        raise NotImplementedError


class AdditiveMultiSourceModel(MultiSourceModel):
    """An AdditiveMultiSourceModel computes the flux from multiple overlapping objects,
    including (host, source pairs) or unresolved sources.

    All rest frame effects are applied to each source, allowing different redshifts
    for each source (for unresolved sources).  The observer frame effects are applied
    to the weighted sum of the sources.

    Note: Each source may have its own sampled (RA, dec) position, which are not
    required to align.

    Attributes
    ----------
    sources : list
        A list of PhysicalModel objects to use in the flux calculation.
    weights : numpy.ndarray, optional
        A length N array of weights to apply to each source. If None, all sources
        will be weighted equally.
    num_sources : int
        The number of sources in the model.

    Parameters
    ----------
    sources : list
        A list of PhysicalModel objects to use in the flux calculation.
    weights : numpy.ndarray, optional
        A length N array of weights to apply to each source. If None, all sources
        will be weighted equally.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(
        self,
        sources,
        weights=None,
        **kwargs,
    ):
        super().__init__(sources, **kwargs)

        if weights is None:
            self.weights = np.full(len(sources), 1.0)
        elif len(weights) != len(sources):
            raise ValueError("Length of weights must match length of sources.")
        else:
            self.weights = weights

    def _evaluate_single(self, times, wavelengths, state, rng_info=None, **kwargs):
        """Evaluate the model and apply the effects for a single, given graph state.
        This function applies redshift, computes the flux density for the object,
        applies rest frames effects, performs the redshift correction (if needed),
        and applies the observer frame effects.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        wavelengths : numpy.ndarray
            A length N array of wavelengths (in angstroms).
        state : GraphState
            An object mapping graph parameters to their values with num_samples=1.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            All the other keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        # Compute the weighted sum of contributions from each source.
        flux_density = np.zeros((len(times), len(wavelengths)))
        for source, weight in zip(self.sources, self.weights, strict=False):
            flux_density += weight * source._evaluate_single(
                times,
                wavelengths,
                state,
                rng_info=rng_info,
                **kwargs,
            )

        # Apply the observer frame effects on the weighted sum of sources.
        params = self.get_local_params(state)
        for effect in self.obs_frame_effects:
            flux_density = effect.apply(
                flux_density,
                times=times,
                wavelengths=wavelengths,
                rng_info=rng_info,
                **params,
            )

        return flux_density


class RandomMultiSourceModel(MultiSourceModel):
    """A RandomMultiSourceModel selects one of its sources at random and
    computes the flux from that source.

    Attributes
    ----------
    source_map : dict
        A dictionary mapping each source name (or index) to a PhysicalModel object.
    num_sources : int
        The number of sources in the model.

    Parameters
    ----------
    sources : list
        A list of PhysicalModel objects to use in the flux calculation.
    weights : numpy.ndarray, optional
        A length N array indicating the relative weight from which to select
        a source at random. If None, all sources will be weighted equally.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(
        self,
        sources,
        weights=None,
        source_names=None,
        **kwargs,
    ):
        super().__init__(sources, **kwargs)

        # Create a parameter to indicate which source was selected.
        source_names = source_names or [src.node_string for src in sources]
        self.source_map = {name: src for name, src in zip(source_names, sources, strict=False)}
        self._sampler_node = GivenValueSampler(source_names, weights=weights)
        self.add_parameter("selected_source", value=self._sampler_node, allow_gradient=False)

    def _evaluate_single(self, times, wavelengths, state, rng_info=None, **kwargs):
        """Evaluate the model and apply the effects for a single, given graph state.
        This function applies redshift, computes the flux density for the object,
        applies rest frames effects, performs the redshift correction (if needed),
        and applies the observer frame effects.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        wavelengths : numpy.ndarray
            A length N array of wavelengths (in angstroms).
        state : GraphState
            An object mapping graph parameters to their values with num_samples=1.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            All the other keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        # Use the model selected by the sampler node to compute the flux density.
        model_name = self.get_param(state, "selected_source")
        flux_density = self.source_map[model_name]._evaluate_single(
            times,
            wavelengths,
            state,
            rng_info=rng_info,
            **kwargs,
        )

        # Apply the observer frame effects on the weighted sum of sources.
        params = self.get_local_params(state)
        for effect in self.obs_frame_effects:
            flux_density = effect.apply(
                flux_density,
                times=times,
                wavelengths=wavelengths,
                rng_info=rng_info,
                **params,
            )

        return flux_density
